# The MIT License (MIT)
#
# Copyright (C) 2015 - Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections import namedtuple
import sys


_MI_VERSION = (1, 0)


class Tags:
    CPU = 'cpu'
    MEMORY = 'memory'
    INTERRUPT = 'interrupt'
    SCHED = 'sched'
    SYSCALL = 'syscall'
    IO = 'io'
    TOP = 'top'
    STATS = 'stats'
    FREQ = 'freq'
    LOG = 'log'
    PERIOD = 'period'


class ColumnDescription:
    def __init__(self, key, title, do_class, unit=None):
        self._key = key
        self._title = title
        self._do_class = do_class
        self._unit = unit

    @property
    def key(self):
        return self._key

    def to_native_object(self):
        obj = {
            'title': self._title,
            'class': self._do_class,
        }

        if self._unit:
            obj['unit'] = self._unit

        return obj


class TableClass:
    def __init__(self, name, title, column_descriptions_tuples=None,
                 inherit=None):
        if column_descriptions_tuples is None:
            column_descriptions_tuples = []

        self._inherit = inherit
        self._name = name
        self._title = title
        self._column_descriptions = []

        for column_descr_tuple in column_descriptions_tuples:
            key = column_descr_tuple[0]
            title = column_descr_tuple[1]
            do_type = column_descr_tuple[2]
            unit = None

            if len(column_descr_tuple) > 3:
                unit = column_descr_tuple[3]

            column_descr = ColumnDescription(key, title, do_type.CLASS, unit)
            self._column_descriptions.append(column_descr)

    @property
    def name(self):
        return self._name

    @property
    def title(self):
        return self._title

    def to_native_object(self):
        obj = {}
        column_descrs = self._column_descriptions
        native_column_descrs = [c.to_native_object() for c in column_descrs]

        if self._inherit is not None:
            obj['inherit'] = self._inherit

        if self._title is not None:
            obj['title'] = self._title

        if native_column_descrs:
            obj['column-descriptions'] = native_column_descrs

        return obj

    def get_column_named_tuple(self):
        keys = [cd.key for cd in self._column_descriptions]

        return namedtuple('Column', keys)


class ResultTable:
    def __init__(self, table_class, begin, end, subtitle=None):
        self._table_class = table_class
        self._column_named_tuple = table_class.get_column_named_tuple()
        self._subtitle = subtitle
        self._timerange = TimeRange(begin, end)
        self._rows = []

    @property
    def table_class(self):
        return self._table_class

    @property
    def timerange(self):
        return self._timerange

    @property
    def title(self):
        return self._table_class.title

    @property
    def subtitle(self):
        return self._subtitle

    def append_row(self, **kwargs):
        row = self._column_named_tuple(**kwargs)
        self._rows.append(row)

    def append_row_tuple(self, row_tuple):
        self._rows.append(row_tuple)

    @property
    def rows(self):
        return self._rows

    def to_native_object(self):
        obj = {
            'class': self._table_class.name,
            'time-range': self._timerange.to_native_object(),
        }
        row_objs = []

        if self._table_class.name:
            if self._subtitle is not None:
                full_title = '{} [{}]'.format(self.title, self._subtitle)
                table_class = TableClass(None, full_title,
                                         inherit=self._table_class.name)
                self._table_class = table_class

        if self._table_class.name is None:
            obj['class'] = self._table_class.to_native_object()

        for row in self._rows:
            row_obj = []

            for cell in row:
                row_obj.append(cell.to_native_object())

            row_objs.append(row_obj)

        obj['data'] = row_objs

        return obj


class _DataObject:
    def to_native_object(self):
        base = {'class': self.CLASS}
        base.update(self._to_native_object())

        return base

    def _to_native_object(self):
        raise NotImplementedError

    def __eq__(self, other):
        # ensure we're comparing the same type first
        if not isinstance(other, self.__class__):
            return False

        # call specific equality method
        return self._eq(other)

    def _eq(self, other):
        raise NotImplementedError


class Empty(_DataObject):
    def to_native_object(self):
        return None

    def _eq(self, other):
        return True


class Unknown(_DataObject):
    CLASS = 'unknown'

    def _to_native_object(self):
        return {}

    def _eq(self, other):
        return True

    def __str__(self):
        return '?'


class _SimpleValue(_DataObject):
    def __init__(self, value):
        self._value = value

    @property
    def value(self):
        return self._value

    def _to_native_object(self):
        return {'value': self._value}

    def __str__(self):
        return str(self._value)

    def _eq(self, other):
        return self.value == other.value


class Boolean(_SimpleValue):
    CLASS = 'bool'


NEG_INF = '-inf'
POS_INF = '+inf'


class Number(_SimpleValue):
    CLASS = 'number'

    def __init__(self, value, low=None, high=None):
        super().__init__(value)
        self._low = low
        self._high = high

    @property
    def low(self):
        return self._low

    @property
    def high(self):
        return self._high

    def _to_native_object(self):
        obj = {}

        if self.value is not None:
            obj['value'] = self.value

        if self._low is not None:
            obj['low'] = self._low

        if self._high is not None:
            obj['high'] = self._high

        return obj

    def _eq(self, other):
        self_tuple = (self.value, self.low, self.high)
        other_tuple = (other.value, other.low, other.high)

        return self_tuple == other_tuple


class String(_SimpleValue):
    CLASS = 'string'


class _SimpleName(_DataObject):
    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        return self._name

    def _to_native_object(self):
        return {'name': self._name}

    def __str__(self):
        return self._name

    def _eq(self, other):
        return self.name == other.name


class Ratio(_SimpleValue):
    CLASS = 'ratio'

    @classmethod
    def from_percentage(cls, value):
        return cls(value / 100)

    def to_percentage(self):
        return self._value * 100


class Timestamp(Number):
    CLASS = 'timestamp'


class Duration(Number):
    CLASS = 'duration'

    @classmethod
    def from_ms(cls, ms):
        return cls(ms * 1000000)

    @classmethod
    def from_us(cls, us):
        return cls(us * 1000)

    def to_ms(self):
        return self._value / 1000000

    def to_us(self):
        return self._value / 1000


class Size(Number):
    CLASS = 'size'


class Bitrate(Number):
    CLASS = 'bitrate'

    @classmethod
    def from_size_duration(cls, size, duration):
        return cls(size * 8 / duration)


class TimeRange(_DataObject):
    CLASS = 'time-range'

    def __init__(self, begin, end):
        self._begin = self._to_timestamp(begin)
        self._end = self._to_timestamp(end)

    @staticmethod
    def _to_timestamp(val):
        if type(val) is int or type(val) is float:
            return Timestamp(val)

        return val

    @property
    def begin(self):
        return self._begin

    @property
    def end(self):
        return self._end

    def _to_native_object(self):
        return {
            'begin': self._begin.to_native_object(),
            'end': self._end.to_native_object()
        }

    def _eq(self, other):
        return (self.begin, self.end) == (other.begin, other.end)


class Syscall(_SimpleName):
    CLASS = 'syscall'


class Process(_DataObject):
    CLASS = 'process'

    def __init__(self, name=None, pid=None, tid=None):
        self._name = name
        self._pid = pid
        self._tid = tid

    @property
    def name(self):
        return self._name

    @property
    def pid(self):
        return self._pid

    @property
    def tid(self):
        return self._tid

    def _to_native_object(self):
        ret_dict = {}

        if self._name is not None:
            ret_dict['name'] = self._name

        if self._pid is not None:
            ret_dict['pid'] = self._pid

        if self._tid is not None:
            ret_dict['tid'] = self._tid

        return ret_dict

    def _eq(self, other):
        self_tuple = (self.name, self.pid, self.tid)
        other_tuple = (other.name, other.pid, other.tid)

        return self_tuple == other_tuple


class Path(_DataObject):
    CLASS = 'path'

    def __init__(self, path):
        self._path = path

    @property
    def path(self):
        return self._path

    def _to_native_object(self):
        return {'path': self._path}

    def _eq(self, other):
        return self.path == other.path


class Fd(_DataObject):
    CLASS = 'fd'

    def __init__(self, fd):
        self._fd = fd

    @property
    def fd(self):
        return self._fd

    def _to_native_object(self):
        return {'fd': self._fd}

    def _eq(self, other):
        return self.fd == other.fd


class Irq(_DataObject):
    CLASS = 'irq'

    def __init__(self, is_hard, nr, name=None):
        self._is_hard = is_hard
        self._nr = nr
        self._name = name

    @property
    def is_hard(self):
        return self._is_hard

    @property
    def nr(self):
        return self._nr

    @property
    def name(self):
        return self._name

    def _to_native_object(self):
        obj = {'hard': self._is_hard, 'nr': self._nr}

        if self._name is not None:
            obj['name'] = self._name

        return obj

    def _eq(self, other):
        self_tuple = (self.is_hard, self.nr, self.name)
        other_tuple = (other.is_hard, other.nr, other.name)

        return self_tuple == other_tuple


class Cpu(_DataObject):
    CLASS = 'cpu'

    def __init__(self, cpu_id):
        self._id = cpu_id

    @property
    def id(self):
        return self._id

    def _to_native_object(self):
        return {'id': self._id}

    def _eq(self, other):
        return self.id == other.id


class Disk(_SimpleName):
    CLASS = 'disk'


class Partition(_SimpleName):
    CLASS = 'part'


class NetIf(_SimpleName):
    CLASS = 'netif'


def get_metadata(version, title, description, authors, url, tags,
                 table_classes):
    t_classes = {t.name: t.to_native_object() for t in table_classes}

    return {
        'mi-version': {
            'major': _MI_VERSION[0],
            'minor': _MI_VERSION[1],
        },
        'version': {
            'major': version.major,
            'minor': version.minor,
            'patch': version.patch,
            'extra': version.extra
        },
        'title': title,
        'description': description,
        'authors': authors,
        'url': url,
        'tags': tags,
        'table-classes': t_classes,
    }


def get_error(message, code=None):
    error = {
        'error-message': message,
    }

    if code is not None:
        error['error-code'] = code

    return error


def get_progress(at=None, msg=None):
    if at is None:
        at = '*'

    add = ''

    if msg is not None:
        add = ' {}'.format(msg)

    return '{}{}'.format(at, add)


def get_version_string():
    return '{}.{}'.format(_MI_VERSION[0], _MI_VERSION[1])


def print_progress(at=None, msg=None):
    print(get_progress(at, msg))
    sys.stdout.flush()



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import sys
import time
from . import mi
from ..common import format_utils


try:
    from progressbar import ETA, Bar, Percentage, ProgressBar
    progressbar_available = True
except ImportError:
    progressbar_available = False


# approximation for the progress bar
_BYTES_PER_EVENT = 30


def get_folder_size(folder):
    total_size = os.path.getsize(folder)
    for item in os.listdir(folder):
        itempath = os.path.join(folder, item)
        if os.path.isfile(itempath):
            total_size += os.path.getsize(itempath)
        elif os.path.isdir(itempath):
            total_size += get_folder_size(itempath)
    return total_size


class _Progress:
    def __init__(self, ts_begin, ts_end, path, use_size=False):
        if ts_begin is None or ts_end is None or use_size:
            size = get_folder_size(path)
            self._maxval = size / _BYTES_PER_EVENT
            self._use_time = False
        else:
            self._maxval = ts_end - ts_begin
            self._ts_begin = ts_begin
            self._ts_end = ts_end
            self._use_time = True

        self._at = 0
        self._event_count = 0
        self._last_event_count_check = 0
        self._last_time_check = time.time()

    def update(self, event):
        self._event_count += 1

        if self._use_time:
            self._at = event.timestamp - self._ts_begin
        else:
            self._at = self._event_count

        if self._at > self._maxval:
            self._at = self._maxval

        if self._event_count - self._last_event_count_check >= 101:
            self._last_event_count_check = self._event_count
            now = time.time()

            if now - self._last_time_check >= .1:
                self._update_progress()
                self._last_time_check = now

    def _update_progress(self):
        pass

    def finalize(self):
        pass


class FancyProgressBar(_Progress):
    def __init__(self, ts_begin, ts_end, path, use_size):
        super().__init__(ts_begin, ts_end, path, use_size)
        self._pbar = None

        if progressbar_available:
            widgets = ['Processing the trace: ', Percentage(), ' ',
                       Bar(marker='#', left='[', right=']'),
                       ' ', ETA(), ' ']  # see docs for other options
            self._pbar = ProgressBar(widgets=widgets,
                                     maxval=self._maxval)
            self._pbar.start()
        else:
            print('Warning: progressbar module not available, '
                  'using --no-progress.', file=sys.stderr)

    def _update_progress(self):
        if self._pbar is None:
            return

        self._pbar.update(self._at)

    def finalize(self):
        if self._pbar is None:
            return

        self._pbar.finish()


class MiProgress(_Progress):
    def __init__(self, ts_begin, ts_end, path, use_size):
        super().__init__(ts_begin, ts_end, path, use_size)

        if self._use_time:
            fmt = 'Starting analysis from {} to {}'
            begin = format_utils.format_timestamp(self._ts_begin)
            end = format_utils.format_timestamp(self._ts_end)
            msg = fmt.format(begin, end)
        else:
            msg = 'Starting analysis: {} estimated events'.format(round(
                self._maxval))

        mi.print_progress(0, msg)

    def _update_progress(self):
        if self._at == self._maxval:
            mi.print_progress(1, 'Done!')
            return

        if self._use_time:
            ts_at = self._at + self._ts_begin
            at_ts = format_utils.format_timestamp(ts_at)
            end = format_utils.format_timestamp(self._ts_end)
            msg = '{}/{}; {} events processed'.format(at_ts, end,
                                                      self._event_count)
        else:
            msg = '{} events processed'.format(self._event_count)

        mi.print_progress(round(self._at / self._maxval, 4), msg)

    def finalize(self):
        mi.print_progress(1, 'Done!')



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#               2015 - Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import operator
from .command import Command
from ..core import memtop
from . import mi
from . import termgraph


class Memtop(Command):
    _DESC = """The memtop command."""
    _ANALYSIS_CLASS = memtop.Memtop
    _MI_TITLE = 'Top memory usage'
    _MI_DESCRIPTION = 'Per-TID top allocated/freed memory'
    _MI_TAGS = [mi.Tags.MEMORY, mi.Tags.TOP]
    _MI_TABLE_CLASS_ALLOCD = 'allocd'
    _MI_TABLE_CLASS_FREED = 'freed'
    _MI_TABLE_CLASS_TOTAL = 'total'
    _MI_TABLE_CLASS_SUMMARY = 'summary'
    _MI_TABLE_CLASSES = [
        (
            _MI_TABLE_CLASS_ALLOCD,
            'Per-TID top allocated memory', [
                ('process', 'Process', mi.Process),
                ('pages', 'Allocated pages', mi.Number, 'pages'),
            ]
        ),
        (
            _MI_TABLE_CLASS_FREED,
            'Per-TID top freed memory', [
                ('process', 'Process', mi.Process),
                ('pages', 'Freed pages', mi.Number, 'pages'),
            ]
        ),
        (
            _MI_TABLE_CLASS_TOTAL,
            'Total allocated/freed memory', [
                ('allocd', 'Total allocated pages', mi.Number, 'pages'),
                ('freed', 'Total freed pages', mi.Number, 'pages'),
            ]
        ),
        (
            _MI_TABLE_CLASS_SUMMARY,
            'Memory usage - summary', [
                ('time_range', 'Time range', mi.TimeRange),
                ('allocd', 'Total allocated pages', mi.Number, 'pages'),
                ('freed', 'Total freed pages', mi.Number, 'pages'),
            ]
        ),
    ]

    def _analysis_tick(self, period_data, end_ns):
        if period_data is None:
            return

        begin_ns = period_data.period.begin_evt.timestamp
        allocd_table = self._get_per_tid_allocd_result_table(period_data,
                                                             begin_ns, end_ns)
        freed_table = self._get_per_tid_freed_result_table(period_data,
                                                           begin_ns, end_ns)
        total_table = self._get_total_result_table(period_data,
                                                   begin_ns, end_ns)

        if self._mi_mode:
            self._mi_append_result_table(allocd_table)
            self._mi_append_result_table(freed_table)
            self._mi_append_result_table(total_table)
        else:
            self._print_date(begin_ns, end_ns)
            self._print_per_tid_allocd(allocd_table)
            self._print_per_tid_freed(freed_table)
            self._print_total(total_table)

    def _create_summary_result_tables(self):
        total_tables = self._mi_get_result_tables(self._MI_TABLE_CLASS_TOTAL)
        begin = total_tables[0].timerange.begin.value
        end = total_tables[-1].timerange.end.value
        summary_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_SUMMARY,
                                         begin, end)

        for total_table in total_tables:
            total_allocd = total_table.rows[0].allocd
            total_freed = total_table.rows[0].freed
            summary_table.append_row(
                time_range=total_table.timerange,
                allocd=total_allocd,
                freed=total_freed,
            )

        self._mi_clear_result_tables()
        self._mi_append_result_table(summary_table)

    def _get_per_tid_attr_result_table(self, period_data, table_class, attr,
                                       begin_ns, end_ns):
        result_table = self._mi_create_result_table(table_class,
                                                    begin_ns, end_ns)
        count = 0

        for tid in sorted(period_data.tids.values(),
                          key=operator.attrgetter(attr),
                          reverse=True):
            result_table.append_row(
                process=mi.Process(tid.comm, tid=tid.tid),
                pages=mi.Number(getattr(tid, attr)),
            )
            count += 1

            if self._args.limit > 0 and count >= self._args.limit:
                break

        return result_table

    def _get_per_tid_allocd_result_table(self, period_data, begin_ns, end_ns):
        return self._get_per_tid_attr_result_table(period_data,
                                                   self._MI_TABLE_CLASS_ALLOCD,
                                                   'allocated_pages',
                                                   begin_ns, end_ns)

    def _get_per_tid_freed_result_table(self, period_data, begin_ns, end_ns):
        return self._get_per_tid_attr_result_table(period_data,
                                                   self._MI_TABLE_CLASS_FREED,
                                                   'freed_pages',
                                                   begin_ns, end_ns)

    def _get_total_result_table(self, period_data, begin_ns, end_ns):
        result_table = self._mi_create_result_table(self._MI_TABLE_CLASS_TOTAL,
                                                    begin_ns, end_ns)
        alloc = 0
        freed = 0

        for tid in period_data.tids.values():
            alloc += tid.allocated_pages
            freed += tid.freed_pages

        result_table.append_row(
            allocd=mi.Number(alloc),
            freed=mi.Number(freed),
        )

        return result_table

    def _print_per_tid_result(self, result_table, title):
        graph = termgraph.BarGraph(
            title=title,
            unit='pages',
            get_value=lambda row: row.pages.value,
            get_label=lambda row: '%s (%d)' % (row.process.name,
                                               row.process.tid),
            label_header='Process',
            data=result_table.rows
        )

        graph.print_graph()

    def _print_per_tid_allocd(self, result_table):
        self._print_per_tid_result(result_table, 'Per-TID Memory Allocations')

    def _print_per_tid_freed(self, result_table):
        self._print_per_tid_result(result_table,
                                   'Per-TID Memory Deallocations')

    def _print_total(self, result_table):
        alloc = result_table.rows[0].allocd.value
        freed = result_table.rows[0].freed.value
        print('\nTotal memory usage:\n- %d pages allocated\n- %d pages freed' %
              (alloc, freed))

    def _add_arguments(self, ap):
        Command._add_proc_filter_args(ap)
        Command._add_top_args(ap)


def _run(mi_mode):
    memtopcmd = Memtop(mi_mode=mi_mode)
    memtopcmd.run()


# entry point (human)
def run():
    _run(mi_mode=False)


# entry point (MI)
def run_mi():
    _run(mi_mode=True)



# The MIT License (MIT)
#
# Copyright (C) 2016 - Julien Desfossez <jdesfossez@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from .analysis import Analysis, PeriodData


class _PeriodData(PeriodData):
    def __init(self):
        self._period_event = None

    @property
    def period_event(self):
        return self._period_event


class PeriodAnalysis(Analysis):
    def __init__(self, state, conf):
        super().__init__(state, conf, {})
        # This is a special case where we keep a global state instead of a
        # per-period state, since we are accumulating statistics about
        # all the periods.
        self._all_period_stats = {}
        self._all_period_list = []
        self._all_total_duration = 0
        self._all_min_duration = None
        self._all_max_duration = None
        # Internal map between currently active periods and their
        # corresponding PeriodEvent object.
        self._current_periods = {}

    def _create_period_data(self):
        return _PeriodData()

    @property
    def all_count(self):
        return len(self._all_period_list)

    @property
    def all_period_stats(self):
        return self._all_period_stats

    @property
    def all_period_list(self):
        return self._all_period_list

    @property
    def all_min_duration(self):
        return self._all_min_duration

    @property
    def all_max_duration(self):
        return self._all_max_duration

    @property
    def all_total_duration(self):
        return self._all_total_duration

    def update_global_stats(self, period_event):
        if self._all_min_duration is None or period_event.duration < \
                self._all_min_duration:
            self._all_min_duration = period_event.duration

        if self._all_max_duration is None or period_event.duration > \
                self._all_max_duration:
            self._all_max_duration = period_event.duration
        self._all_total_duration += period_event.duration

    # beginning of a new period
    def _begin_period_cb(self, period_data):
        # Only track real periods, not the dummy ones created
        # when no --period argument was passed.
        if period_data.period.definition is None:
            return

        period = period_data.period
        definition = period.definition

        if definition.name not in self._all_period_stats:
            if definition.name is None:
                name = ""
            else:
                name = definition.name
            self._all_period_stats[name] = \
                PeriodStats.new_from_period(period_data.period)

        if period.parent is not None:
            parent = self._current_periods[period.parent]
        else:
            parent = None

        period_data._period_event = PeriodEvent(
            period.begin_evt.timestamp, definition.name, parent)

        self._all_period_list.append(period_data._period_event)
        self._current_periods[period] = period_data._period_event

    def _end_period_cb(self, period_data, completed,
                       begin_captures, end_captures):
        period = period_data.period

        if period.definition is None:
            return

        if completed is False:
            # We should eventually warn the user here or keep
            # the event as uncomplete or in a separate table.
            self._all_period_list.remove(period_data._period_event)
            return

        if period.definition.name is None:
            name = ""
        else:
            name = period.definition.name

        period_data._period_event.finish(
            self.last_event_ts, begin_captures, end_captures)
        self._all_period_stats[name].update_stats(
            period_data._period_event)
        self.update_global_stats(period_data._period_event)

        if period.parent is not None:
            parent = self._current_periods[period.parent]
            parent.add_child(period_data._period_event)

        del self._current_periods[period]


class PeriodStats():
    def __init__(self, name):
        self.name = name
        self.period_list = []
        self.min_duration = None
        self.max_duration = None
        self.total_duration = 0

    @classmethod
    def new_from_period(cls, period):
        if period.definition.name is None:
            return cls("")
        return cls(period.definition.name)

    @property
    def count(self):
        return len(self.period_list)

    def update_stats(self, period_event):
        if self.min_duration is None or period_event.duration < \
                self.min_duration:
            self.min_duration = period_event.duration

        if self.max_duration is None or period_event.duration > \
                self.max_duration:
            self.max_duration = period_event.duration
        self.total_duration += period_event.duration
        self.period_list.append(period_event)


class PeriodEvent():
    def __init__(self, start_ts, name, parent):
        self._start_ts = start_ts
        self._name = name
        self._parent = parent
        self._end_ts = None
        self._begin_captures = None
        self._end_captures = None
        # Only during the aggregation phase, store the list
        # of children we want to output.
        self._children = []

    @property
    def start_ts(self):
        return self._start_ts

    @property
    def end_ts(self):
        return self._end_ts

    @property
    def name(self):
        if self._name is None:
            return ""
        return self._name

    @property
    def duration(self):
        return self._end_ts - self._start_ts

    @property
    def begin_captures(self):
        return str(self._begin_captures)

    @property
    def end_captures(self):
        return str(self._end_captures)

    def filtered_captures(self, period_group_by):
        # List of tuple (field, value) for all the captured fields
        # present in the _period_group_by dict.
        _captures = []
        if self._name not in period_group_by.keys():
            return _captures
        if self._begin_captures is not None:
            for c in sorted(self._begin_captures.keys()):
                if c in period_group_by[self._name]:
                    _captures.append(('%s.%s' % (self._name, c),
                                     self._begin_captures[c]))
        if self._end_captures is not None:
            for c in sorted(self._end_captures.keys()):
                if c in period_group_by[self._name]:
                    _captures.append(('%s.%s' % (self._name, c),
                                     self._end_captures[c]))
        return _captures

    def full_captures(self):
        _captures = []
        if self._begin_captures is not None:
            for c in self._begin_captures.keys():
                _captures.append(('%s.%s' % (self._name, c),
                                 self._begin_captures[c]))
        if self._end_captures is not None:
            for c in self._end_captures.keys():
                _captures.append(('%s.%s' % (self._name, c),
                                 self._end_captures[c]))
        return _captures

    @property
    def parent(self):
        return self._parent

    @property
    def children(self):
        return self._children

    def finish(self, end_ts, begin_captures, end_captures):
        self._end_ts = end_ts
        self._begin_captures = begin_captures
        self._end_captures = end_captures

    def add_child(self, child_period_event):
        self._children.append(child_period_event)



# The MIT License (MIT)
#
# Copyright (C) 2016 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import unittest
from lttnganalyses.core import stats
from lttnganalyses.common import format_utils
from .utils import TimezoneUtils


class TestFormatSize(unittest.TestCase):
    def test_negative(self):
        self.assertRaises(ValueError, format_utils.format_size, -1)

    def test_zero(self):
        result = format_utils.format_size(0)
        result_decimal = format_utils.format_size(0, binary_prefix=False)

        self.assertEqual(result, '0   B')
        self.assertEqual(result_decimal, '0  B')

    def test_huge(self):
        # 2000 YiB or 2475.88 YB
        huge_value = 2417851639229258349412352000
        result = format_utils.format_size(huge_value)
        result_decimal = format_utils.format_size(huge_value,
                                                  binary_prefix=False)

        self.assertEqual(result, '2000.00 YiB')
        self.assertEqual(result_decimal, '2417.85 YB')

    def test_reasonable(self):
        # 2 GB or 1.86 GiB
        reasonable_value = 2000000000
        result = format_utils.format_size(reasonable_value)
        result_decimal = format_utils.format_size(reasonable_value,
                                                  binary_prefix=False)

        self.assertEqual(result, '1.86 GiB')
        self.assertEqual(result_decimal, '2.00 GB')


class TestFormatPrioList(unittest.TestCase):
    def test_empty(self):
        prio_list = []
        result = format_utils.format_prio_list(prio_list)

        self.assertEqual(result, '[]')

    def test_one_prio(self):
        prio_list = [stats.PrioEvent(0, 0)]
        result = format_utils.format_prio_list(prio_list)

        self.assertEqual(result, '[0]')

    def test_multiple_prios(self):
        prio_list = [stats.PrioEvent(0, 0), stats.PrioEvent(0, 1)]
        result = format_utils.format_prio_list(prio_list)

        self.assertEqual(result, '[0, 1]')

    def test_repeated_prio(self):
        prio_list = [stats.PrioEvent(0, 0), stats.PrioEvent(0, 0)]
        result = format_utils.format_prio_list(prio_list)

        self.assertEqual(result, '[0 (2 times)]')

    def test_repeated_prios(self):
        prio_list = [
            stats.PrioEvent(0, 0), stats.PrioEvent(0, 1),
            stats.PrioEvent(0, 0), stats.PrioEvent(0, 1)
        ]
        result = format_utils.format_prio_list(prio_list)

        self.assertEqual(result, '[0 (2 times), 1 (2 times)]')


class TestFormatTimestamp(unittest.TestCase):
    # This may or may not be the time of the Linux 0.0.1 announcement.
    ARBITRARY_TIMESTAMP = 683153828123456789

    def setUp(self):
        self.tz_utils = TimezoneUtils()
        self.tz_utils.set_up_timezone()

    def tearDown(self):
        self.tz_utils.tear_down_timezone()

    def test_time(self):
        result = format_utils.format_timestamp(self.ARBITRARY_TIMESTAMP)
        result_gmt = format_utils.format_timestamp(
            self.ARBITRARY_TIMESTAMP, gmt=True
        )

        self.assertEqual(result, '16:57:08.123456789')
        self.assertEqual(result_gmt, '20:57:08.123456789')

    def test_date(self):
        result = format_utils.format_timestamp(
            self.ARBITRARY_TIMESTAMP, print_date=True
        )
        result_gmt = format_utils.format_timestamp(
            self.ARBITRARY_TIMESTAMP, print_date=True, gmt=True
        )

        self.assertEqual(result, '1991-08-25 16:57:08.123456789')
        self.assertEqual(result_gmt, '1991-08-25 20:57:08.123456789')

    def test_negative(self):
        # Make sure the time module handles pre-epoch dates correctly
        result = format_utils.format_timestamp(
            -self.ARBITRARY_TIMESTAMP, print_date=True
        )
        result_gmt = format_utils.format_timestamp(
            -self.ARBITRARY_TIMESTAMP, print_date=True, gmt=True
        )

        self.assertEqual(result, '1948-05-08 23:02:51.876543211')
        self.assertEqual(result_gmt, '1948-05-09 03:02:51.876543211')


class TestFormatTimeRange(unittest.TestCase):
    BEGIN_TS = 683153828123456789
    # 1 hour later
    END_TS = 683157428123456789

    def _mock_format_timestamp(self, timestamp, print_date, gmt):
        date_str = '1991-08-25 '

        if timestamp == TestFormatTimeRange.BEGIN_TS:
            if gmt:
                time_str = '20:57:08.123456789'
            else:
                time_str = '16:57:08.123456789'
        elif timestamp == TestFormatTimeRange.END_TS:
            if gmt:
                time_str = '21:57:08.123456789'
            else:
                time_str = '17:57:08.123456789'

        if print_date:
            return date_str + time_str
        else:
            return time_str

    def setUp(self):
        self._original_format_timestamp = format_utils.format_timestamp
        format_utils.format_timestamp = self._mock_format_timestamp

    def tearDown(self):
        format_utils.format_timestamp = self._original_format_timestamp

    def test_time_only(self):
        result = format_utils.format_time_range(
            self.BEGIN_TS, self.END_TS
        )
        result_gmt = format_utils.format_time_range(
            self.BEGIN_TS, self.END_TS, gmt=True
        )

        self.assertEqual(result,
                         '[16:57:08.123456789, 17:57:08.123456789]')
        self.assertEqual(result_gmt,
                         '[20:57:08.123456789, 21:57:08.123456789]')

    def test_print_date(self):
        result = format_utils.format_time_range(
            self.BEGIN_TS, self.END_TS, print_date=True
        )
        result_gmt = format_utils.format_time_range(
            self.BEGIN_TS, self.END_TS, print_date=True, gmt=True
        )

        self.assertEqual(
            result,
            '[1991-08-25 16:57:08.123456789, 1991-08-25 17:57:08.123456789]'
        )
        self.assertEqual(
            result_gmt,
            '[1991-08-25 20:57:08.123456789, 1991-08-25 21:57:08.123456789]'
        )


class TestFormatIpv4(unittest.TestCase):
    IP_INTEGER = 0x7f000001
    IP_SEQUENCE = [127, 0, 0, 1]

    def test_integer(self):
        result = format_utils.format_ipv4(self.IP_INTEGER)

        self.assertEqual(result, '127.0.0.1')

    def test_sequence(self):
        result = format_utils.format_ipv4(self.IP_SEQUENCE)

        self.assertEqual(result, '127.0.0.1')

    def test_with_port(self):
        result = format_utils.format_ipv4(self.IP_SEQUENCE, port=8080)

        self.assertEqual(result, '127.0.0.1:8080')



#!/usr/bin/env python3
#
# Copyright (C) 2015 - Michael Jeanson <mjeanson@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""LTTnganalyses setup script"""


import shutil
import sys
from setuptools import setup
import versioneer

if sys.version_info[0:2] < (3, 4):
    raise RuntimeError("Python version >= 3.4 required.")

if 'install' in sys.argv:
    if shutil.which('babeltrace') is None:
        print('lttnganalysescli needs the babeltrace executable.\n'
              'See https://www.efficios.com/babeltrace for more info.',
              file=sys.stderr)
        sys.exit(1)

    try:
        __import__('babeltrace')
    except ImportError:
        print('lttnganalysescli needs the babeltrace python bindings.\n'
              'See https://www.efficios.com/babeltrace for more info.',
              file=sys.stderr)
        sys.exit(1)


def read_file(filename):
    """Read all contents of ``filename``."""
    with open(filename, encoding='utf-8') as source:
        return source.read()


setup(
    name='lttnganalyses',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),

    description='LTTng analyses',
    long_description=read_file('README.rst'),

    url='https://github.com/lttng/lttng-analyses',

    author='Julien Desfossez',
    author_email='jdesfossez@efficios.com',

    license='MIT',

    classifiers=[
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: System :: Monitoring',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.4',
    ],

    keywords='lttng tracing',

    packages=[
        'lttnganalyses',
        'lttnganalyses.common',
        'lttnganalyses.core',
        'lttnganalyses.cli',
        'lttnganalyses.linuxautomaton'
        ],

    entry_points={
        'console_scripts': [
            # human-readable output
            'lttng-cputop = lttnganalyses.cli.cputop:run',
            'lttng-iolatencyfreq = lttnganalyses.cli.io:runfreq',
            'lttng-iolatencystats = lttnganalyses.cli.io:runstats',
            'lttng-iolatencytop = lttnganalyses.cli.io:runlatencytop',
            'lttng-iolog = lttnganalyses.cli.io:runlog',
            'lttng-iousagetop = lttnganalyses.cli.io:runusage',
            'lttng-irqfreq = lttnganalyses.cli.irq:runfreq',
            'lttng-irqlog = lttnganalyses.cli.irq:runlog',
            'lttng-irqstats = lttnganalyses.cli.irq:runstats',
            'lttng-memtop = lttnganalyses.cli.memtop:run',
            'lttng-syscallstats = lttnganalyses.cli.syscallstats:run',
            'lttng-schedlog = lttnganalyses.cli.sched:runlog',
            'lttng-schedtop = lttnganalyses.cli.sched:runtop',
            'lttng-schedstats = lttnganalyses.cli.sched:runstats',
            'lttng-schedfreq = lttnganalyses.cli.sched:runfreq',
            'lttng-periodlog = lttnganalyses.cli.periods:runlog',
            'lttng-periodtop = lttnganalyses.cli.periods:runtop',
            'lttng-periodstats = lttnganalyses.cli.periods:runstats',
            'lttng-periodfreq = lttnganalyses.cli.periods:runfreq',

            # MI mode
            'lttng-cputop-mi = lttnganalyses.cli.cputop:run_mi',
            'lttng-memtop-mi = lttnganalyses.cli.memtop:run_mi',
            'lttng-syscallstats-mi = lttnganalyses.cli.syscallstats:run_mi',
            'lttng-irqfreq-mi = lttnganalyses.cli.irq:runfreq_mi',
            'lttng-irqlog-mi = lttnganalyses.cli.irq:runlog_mi',
            'lttng-irqstats-mi = lttnganalyses.cli.irq:runstats_mi',
            'lttng-iolatencyfreq-mi = lttnganalyses.cli.io:runfreq_mi',
            'lttng-iolatencystats-mi = lttnganalyses.cli.io:runstats_mi',
            'lttng-iolatencytop-mi = lttnganalyses.cli.io:runlatencytop_mi',
            'lttng-iolog-mi = lttnganalyses.cli.io:runlog_mi',
            'lttng-iousagetop-mi = lttnganalyses.cli.io:runusage_mi',
            'lttng-schedlog-mi = lttnganalyses.cli.sched:runlog_mi',
            'lttng-schedtop-mi = lttnganalyses.cli.sched:runtop_mi',
            'lttng-schedstats-mi = lttnganalyses.cli.sched:runstats_mi',
            'lttng-schedfreq-mi = lttnganalyses.cli.sched:runfreq_mi',
            'lttng-periodlog-mi = lttnganalyses.cli.periods:runlog_mi',
            'lttng-periodtop-mi = lttnganalyses.cli.periods:runtop_mi',
            'lttng-periodstats-mi = lttnganalyses.cli.periods:runstats_mi',
            'lttng-periodfreq-mi = lttnganalyses.cli.periods:runfreq_mi',
        ],
    },

    scripts=[
        'lttng-analyses-record',
        'lttng-track-process'
    ],

    install_requires=[
        'pyparsing',
    ],

    extras_require={
        'progressbar': ["progressbar"]
    },

    test_suite='tests',
)



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#               2015 - Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import collections
import operator
import statistics
import sys
from . import mi
from . import termgraph
from ..core import io
from ..common import format_utils
from .command import Command


_UsageTables = collections.namedtuple('_UsageTables', [
    'per_proc_read',
    'per_proc_write',
    'per_file_read',
    'per_file_write',
    'per_proc_block_read',
    'per_proc_block_write',
    'per_disk_sector',
    'per_disk_request',
    'per_disk_rtps',
    'per_netif_recv',
    'per_netif_send',
])


class IoAnalysisCommand(Command):
    _DESC = """The I/O command."""
    _ANALYSIS_CLASS = io.IoAnalysis
    _MI_TITLE = 'I/O analysis'
    _MI_DESCRIPTION = 'System call/disk latency statistics, system call ' + \
                      'latency distribution, system call top latencies, ' + \
                      'I/O usage top, and I/O operations log'
    _MI_TAGS = [
        mi.Tags.IO,
        mi.Tags.SYSCALL,
        mi.Tags.STATS,
        mi.Tags.FREQ,
        mi.Tags.LOG,
        mi.Tags.TOP,
    ]
    _MI_TABLE_CLASS_SYSCALL_LATENCY_STATS = 'syscall-latency-stats'
    _MI_TABLE_CLASS_PART_LATENCY_STATS = 'disk-latency-stats'
    _MI_TABLE_CLASS_FREQ = 'freq'
    _MI_TABLE_CLASS_TOP_SYSCALL = 'top-syscall'
    _MI_TABLE_CLASS_LOG = 'log'
    _MI_TABLE_CLASS_PER_PROCESS_TOP = 'per-process-top'
    _MI_TABLE_CLASS_PER_FILE_TOP = 'per-file-top'
    _MI_TABLE_CLASS_PER_PROCESS_TOP_BLOCK = 'per-process-top-block'
    _MI_TABLE_CLASS_PER_DISK_TOP_SECTOR = 'per-disk-top-sector'
    _MI_TABLE_CLASS_PER_DISK_TOP_REQUEST = 'per-disk-top-request'
    _MI_TABLE_CLASS_PER_DISK_TOP_RTPS = 'per-disk-top-rps'
    _MI_TABLE_CLASS_PER_NETIF_TOP = 'per-netif-top'
    _MI_TABLE_CLASSES = [
        (
            _MI_TABLE_CLASS_SYSCALL_LATENCY_STATS,
            'System call latency statistics', [
                ('obj', 'System call category', mi.String),
                ('count', 'Call count', mi.Number, 'calls'),
                ('min_latency', 'Minimum call latency', mi.Duration),
                ('avg_latency', 'Average call latency', mi.Duration),
                ('max_latency', 'Maximum call latency', mi.Duration),
                ('stdev_latency', 'System call latency standard deviation',
                 mi.Duration),
            ]
        ),
        (
            _MI_TABLE_CLASS_PART_LATENCY_STATS,
            'Partition latency statistics', [
                ('obj', 'Partition', mi.Disk),
                ('count', 'Access count', mi.Number, 'accesses'),
                ('min_latency', 'Minimum access latency', mi.Duration),
                ('avg_latency', 'Average access latency', mi.Duration),
                ('max_latency', 'Maximum access latency', mi.Duration),
                ('stdev_latency', 'System access latency standard deviation',
                 mi.Duration),
            ]
        ),
        (
            _MI_TABLE_CLASS_FREQ,
            'I/O request latency distribution', [
                ('latency_lower', 'Latency (lower bound)', mi.Duration),
                ('latency_upper', 'Latency (upper bound)', mi.Duration),
                ('count', 'Request count', mi.Number, 'requests'),
            ]
        ),
        (
            _MI_TABLE_CLASS_TOP_SYSCALL,
            'Top system call latencies', [
                ('time_range', 'Call time range', mi.TimeRange),
                ('out_of_range', 'System call out of range?', mi.Boolean),
                ('duration', 'Call duration', mi.Duration),
                ('syscall', 'System call', mi.Syscall),
                ('size', 'Read/write size', mi.Size),
                ('process', 'Process', mi.Process),
                ('path', 'File path', mi.Path),
                ('fd', 'File descriptor', mi.Fd),
            ]
        ),
        (
            _MI_TABLE_CLASS_LOG,
            'I/O operations log', [
                ('time_range', 'Call time range', mi.TimeRange),
                ('out_of_range', 'System call out of range?', mi.Boolean),
                ('duration', 'Call duration', mi.Duration),
                ('syscall', 'System call', mi.Syscall),
                ('size', 'Read/write size', mi.Size),
                ('process', 'Process', mi.Process),
                ('path', 'File path', mi.Path),
                ('fd', 'File descriptor', mi.Fd),
            ]
        ),
        (
            _MI_TABLE_CLASS_PER_PROCESS_TOP,
            'Per-process top I/O operations', [
                ('process', 'Process', mi.Process),
                ('size', 'Total operations size', mi.Size),
                ('disk_size', 'Disk operations size', mi.Size),
                ('net_size', 'Network operations size', mi.Size),
                ('unknown_size', 'Unknown operations size', mi.Size),
            ]
        ),
        (
            _MI_TABLE_CLASS_PER_FILE_TOP,
            'Per-file top I/O operations', [
                ('path', 'File path/info', mi.Path),
                ('size', 'Operations size', mi.Size),
                ('fd_owners', 'File descriptor owners', mi.String),
            ]
        ),
        (
            _MI_TABLE_CLASS_PER_PROCESS_TOP_BLOCK,
            'Per-process top block I/O operations', [
                ('process', 'Process', mi.Process),
                ('size', 'Operations size', mi.Size),
            ]
        ),
        (
            _MI_TABLE_CLASS_PER_DISK_TOP_SECTOR,
            'Per-disk top sector I/O operations', [
                ('disk', 'Disk', mi.Disk),
                ('count', 'Sector count', mi.Number, 'sectors'),
            ]
        ),
        (
            _MI_TABLE_CLASS_PER_DISK_TOP_REQUEST,
            'Per-disk top I/O requests', [
                ('disk', 'Disk', mi.Disk),
                ('count', 'Request count', mi.Number, 'I/O requests'),
            ]
        ),
        (
            _MI_TABLE_CLASS_PER_DISK_TOP_RTPS,
            'Per-disk top I/O request time/sector', [
                ('disk', 'Disk', mi.Disk),
                ('rtps', 'Request time/sector', mi.Duration),
            ]
        ),
        (
            _MI_TABLE_CLASS_PER_NETIF_TOP,
            'Per-network interface top I/O operations', [
                ('netif', 'Network interface', mi.NetIf),
                ('size', 'Operations size', mi.Size),
            ]
        ),
    ]
    _LATENCY_STATS_FORMAT = '{:<14} {:>14} {:>14} {:>14} {:>14} {:>14}'
    _SECTION_SEPARATOR_STRING = '-' * 89

    def _analysis_tick(self, period_data, end_ns):
        if period_data is None:
            return

        begin_ns = period_data.period.begin_evt.timestamp
        syscall_latency_stats_table = None
        disk_latency_stats_table = None
        freq_tables = None
        top_tables = None
        log_table = None
        usage_tables = None

        if self._args.stats:
            syscall_latency_stats_table, disk_latency_stats_table = \
                self._get_latency_stats_result_tables(period_data,
                                                      begin_ns, end_ns)

        if self._args.freq:
            freq_tables = self._get_freq_result_tables(period_data, begin_ns,
                                                       end_ns)

        if self._args.usage:
            usage_tables = self._get_usage_result_tables(period_data, begin_ns,
                                                         end_ns)

        if self._args.top:
            top_tables = self._get_top_result_tables(period_data,
                                                     begin_ns, end_ns)

        if self._args.log:
            log_table = self._get_log_result_table(period_data,
                                                   begin_ns, end_ns)

        if self._mi_mode:
            self._mi_append_result_tables([
                log_table,
                syscall_latency_stats_table,
                disk_latency_stats_table,
            ])
            self._mi_append_result_tables(top_tables)
            self._mi_append_result_tables(usage_tables)
            self._mi_append_result_tables(freq_tables)
        else:
            self._print_date(begin_ns, end_ns)

            if self._args.usage:
                self._print_usage(usage_tables)

            if self._args.stats:
                self._print_latency_stats(syscall_latency_stats_table,
                                          disk_latency_stats_table)

            if self._args.top:
                self._print_top(top_tables)

            if self._args.freq:
                self._print_freq(freq_tables)

            if self._args.log:
                self._print_log(log_table)

    def _create_summary_result_tables(self):
        # TODO: create a summary table here
        self._mi_clear_result_tables()

    # Filter predicates
    def _filter_size(self, size):
        if size is None:
            return True
        if self._args.maxsize is not None and size > self._args.maxsize:
            return False
        if self._args.minsize is not None and size < self._args.minsize:
            return False
        return True

    def _filter_latency(self, duration):
        if self._args.max is not None and duration > self._args.max:
            return False
        if self._args.min is not None and duration < self._args.min:
            return False
        return True

    def _filter_time_range(self, begin, end):
        # Note: we only want to return False only when a request has
        # ended and is completely outside the timerange (i.e. begun
        # after the end of the time range).
        return not (
            self._analysis_conf.begin_ts and self._analysis_conf.end_ts and
            end and begin > self._analysis_conf.end_ts
        )

    def _filter_io_request(self, io_rq):
        return self._filter_size(io_rq.size) and \
            self._filter_latency(io_rq.duration) and \
            self._filter_time_range(io_rq.begin_ts, io_rq.end_ts)

    def _is_io_rq_out_of_range(self, io_rq):
        return (
            self._analysis_conf.begin_ts and
            io_rq.begin_ts < self._analysis_conf.begin_ts or
            self._analysis_conf.end_ts and
            io_rq.end_ts > self._analysis_conf.end_ts
        )

    def _append_per_proc_read_usage_row(self, period_data, proc_stats,
                                        result_table):
        result_table.append_row(
            process=mi.Process(proc_stats.comm, pid=proc_stats.pid,
                               tid=proc_stats.tid),
            size=mi.Size(proc_stats.total_read),
            disk_size=mi.Size(proc_stats.disk_io.read),
            net_size=mi.Size(proc_stats.net_io.read),
            unknown_size=mi.Size(proc_stats.unk_io.read),
        )

        return True

    def _append_per_proc_write_usage_row(self, period_data, proc_stats,
                                         result_table):
        result_table.append_row(
            process=mi.Process(proc_stats.comm, pid=proc_stats.pid,
                               tid=proc_stats.tid),
            size=mi.Size(proc_stats.total_write),
            disk_size=mi.Size(proc_stats.disk_io.write),
            net_size=mi.Size(proc_stats.net_io.write),
            unknown_size=mi.Size(proc_stats.unk_io.write),
        )

        return True

    def _append_per_proc_block_read_usage_row(self, period_data, proc_stats,
                                              result_table):
        if proc_stats.block_io.read == 0:
            return False

        if proc_stats.comm:
            proc_name = proc_stats.comm
        else:
            proc_name = None

        result_table.append_row(
            process=mi.Process(proc_name, pid=proc_stats.pid,
                               tid=proc_stats.tid),
            size=mi.Size(proc_stats.block_io.read),
        )

        return True

    def _append_per_proc_block_write_usage_row(self, period_data, proc_stats,
                                               result_table):
        if proc_stats.block_io.write == 0:
            return False

        if proc_stats.comm:
            proc_name = proc_stats.comm
        else:
            proc_name = None

        result_table.append_row(
            process=mi.Process(proc_name, pid=proc_stats.pid,
                               tid=proc_stats.tid),
            size=mi.Size(proc_stats.block_io.write),
        )

        return True

    def _append_disk_sector_usage_row(self, period_data, disk_stats,
                                      result_table):
        if disk_stats.total_rq_sectors == 0:
            return None

        result_table.append_row(
            disk=mi.Disk(disk_stats.diskname),
            count=mi.Number(disk_stats.total_rq_sectors),
        )

        return True

    def _append_disk_request_usage_row(self, period_data, disk_stats,
                                       result_table):
        if disk_stats.rq_count == 0:
            return False

        result_table.append_row(
            disk=mi.Disk(disk_stats.diskname),
            count=mi.Number(disk_stats.rq_count),
        )

        return True

    def _append_disk_rtps_usage_row(self, period_data, disk_stats,
                                    result_table):
        if disk_stats.rq_count == 0:
            return False

        avg_latency = (disk_stats.total_rq_duration / disk_stats.rq_count)
        result_table.append_row(
            disk=mi.Disk(disk_stats.diskname),
            rtps=mi.Duration(avg_latency),
        )

        return True

    def _append_netif_recv_usage_row(self, period_data, netif_stats,
                                     result_table):
        result_table.append_row(
            netif=mi.NetIf(netif_stats.name),
            size=mi.Size(netif_stats.recv_bytes)
        )

        return True

    def _append_netif_send_usage_row(self, period_data, netif_stats,
                                     result_table):
        result_table.append_row(
            netif=mi.NetIf(netif_stats.name),
            size=mi.Size(netif_stats.sent_bytes)
        )

        return True

    def _get_file_stats_fd_owners_str(self, period_data, file_stats):
        fd_by_pid_str = ''

        for pid, fd in file_stats.fd_by_pid.items():
            comm = period_data.tids[pid].comm
            fd_by_pid_str += 'fd %d in %s (%s) ' % (fd, comm, pid)

        return fd_by_pid_str

    def _append_file_read_usage_row(self, period_data, file_stats,
                                    result_table):
        if file_stats.io.read == 0:
            return False

        fd_owners = self._get_file_stats_fd_owners_str(period_data, file_stats)
        result_table.append_row(
            path=mi.Path(file_stats.filename),
            size=mi.Size(file_stats.io.read),
            fd_owners=mi.String(fd_owners),
        )

        return True

    def _append_file_write_usage_row(self, period_data, file_stats,
                                     result_table):
        if file_stats.io.write == 0:
            return False

        fd_owners = self._get_file_stats_fd_owners_str(period_data, file_stats)
        result_table.append_row(
            path=mi.Path(file_stats.filename),
            size=mi.Size(file_stats.io.write),
            fd_owners=mi.String(fd_owners),
        )

        return True

    def _fill_usage_result_table(self, period_data, input_list, append_row_cb,
                                 result_table):
        count = 0
        limit = self._args.limit

        for elem in input_list:
            if append_row_cb(period_data, elem, result_table):
                count += 1

                if limit is not None and count >= limit:
                    break

    def _fill_per_process_read_usage_result_table(self, period_data,
                                                  result_table):
        input_list = sorted(period_data.tids.values(),
                            key=operator.attrgetter('total_read'),
                            reverse=True)
        self._fill_usage_result_table(period_data, input_list,
                                      self._append_per_proc_read_usage_row,
                                      result_table)

    def _fill_per_process_write_usage_result_table(self, period_data,
                                                   result_table):
        input_list = sorted(period_data.tids.values(),
                            key=operator.attrgetter('total_write'),
                            reverse=True)
        self._fill_usage_result_table(period_data, input_list,
                                      self._append_per_proc_write_usage_row,
                                      result_table)

    def _fill_per_process_block_read_usage_result_table(self, period_data,
                                                        result_table):
        input_list = sorted(period_data.tids.values(),
                            key=operator.attrgetter('block_io.read'),
                            reverse=True)
        self._fill_usage_result_table(
            period_data, input_list,
            self._append_per_proc_block_read_usage_row,
            result_table)

    def _fill_per_process_block_write_usage_result_table(self, period_data,
                                                         result_table):
        input_list = sorted(period_data.tids.values(),
                            key=operator.attrgetter('block_io.write'),
                            reverse=True)
        self._fill_usage_result_table(
            period_data, input_list,
            self._append_per_proc_block_write_usage_row, result_table)

    def _fill_disk_sector_usage_result_table(self, period_data, result_table):
        input_list = sorted(period_data.disks.values(),
                            key=operator.attrgetter('total_rq_sectors'),
                            reverse=True)
        self._fill_usage_result_table(period_data, input_list,
                                      self._append_disk_sector_usage_row,
                                      result_table)

    def _fill_disk_request_usage_result_table(self, period_data, result_table):
        input_list = sorted(period_data.disks.values(),
                            key=operator.attrgetter('rq_count'),
                            reverse=True)
        self._fill_usage_result_table(period_data, input_list,
                                      self._append_disk_request_usage_row,
                                      result_table)

    def _fill_disk_rtps_usage_result_table(self, period_data, result_table):
        input_list = period_data.disks.values()
        self._fill_usage_result_table(period_data, input_list,
                                      self._append_disk_rtps_usage_row,
                                      result_table)

    def _fill_netif_recv_usage_result_table(self, period_data, result_table):
        input_list = sorted(period_data.ifaces.values(),
                            key=operator.attrgetter('recv_bytes'),
                            reverse=True)
        self._fill_usage_result_table(period_data, input_list,
                                      self._append_netif_recv_usage_row,
                                      result_table)

    def _fill_netif_send_usage_result_table(self, period_data, result_table):
        input_list = sorted(period_data.ifaces.values(),
                            key=operator.attrgetter('sent_bytes'),
                            reverse=True)
        self._fill_usage_result_table(period_data, input_list,
                                      self._append_netif_send_usage_row,
                                      result_table)

    def _fill_file_read_usage_result_table(self, period_data, files,
                                           result_table):
        input_list = sorted(files.values(),
                            key=lambda file_stats: file_stats.io.read,
                            reverse=True)
        self._fill_usage_result_table(period_data, input_list,
                                      self._append_file_read_usage_row,
                                      result_table)

    def _fill_file_write_usage_result_table(self, period_data, files,
                                            result_table):
        input_list = sorted(files.values(),
                            key=lambda file_stats: file_stats.io.write,
                            reverse=True)
        self._fill_usage_result_table(period_data, input_list,
                                      self._append_file_write_usage_row,
                                      result_table)

    def _fill_file_usage_result_tables(self, period_data, read_table,
                                       write_table):
        files = self._analysis.get_files_stats(period_data)
        self._fill_file_read_usage_result_table(period_data, files,
                                                read_table)
        self._fill_file_write_usage_result_table(period_data, files,
                                                 write_table)

    def _get_usage_result_tables(self, period_data, begin, end):
        # create result tables
        per_proc_read_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_PROCESS_TOP, begin, end, 'read')
        per_proc_write_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_PROCESS_TOP, begin, end, 'written')
        per_file_read_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_FILE_TOP, begin, end, 'read')
        per_file_write_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_FILE_TOP, begin, end, 'written')
        per_proc_block_read_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_PROCESS_TOP_BLOCK, begin, end, 'read')
        per_proc_block_write_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_PROCESS_TOP_BLOCK, begin, end, 'written')
        per_disk_sector_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_DISK_TOP_SECTOR, begin, end)
        per_disk_request_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_DISK_TOP_REQUEST, begin, end)
        per_disk_rtps_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_DISK_TOP_RTPS, begin, end)
        per_netif_recv_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_NETIF_TOP, begin, end, 'received')
        per_netif_send_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PER_NETIF_TOP, begin, end, 'sent')

        # fill result tables
        self._fill_per_process_read_usage_result_table(period_data,
                                                       per_proc_read_table)
        self._fill_per_process_write_usage_result_table(period_data,
                                                        per_proc_write_table)
        self._fill_file_usage_result_tables(period_data, per_file_read_table,
                                            per_file_write_table)
        self._fill_per_process_block_read_usage_result_table(
            period_data, per_proc_block_read_table)
        self._fill_per_process_block_write_usage_result_table(
            period_data, per_proc_block_write_table)
        self._fill_disk_sector_usage_result_table(period_data,
                                                  per_disk_sector_table)
        self._fill_disk_request_usage_result_table(period_data,
                                                   per_disk_request_table)
        self._fill_disk_rtps_usage_result_table(period_data,
                                                per_disk_rtps_table)
        self._fill_netif_recv_usage_result_table(period_data,
                                                 per_netif_recv_table)
        self._fill_netif_send_usage_result_table(period_data,
                                                 per_netif_send_table)

        return _UsageTables(
            per_proc_read=per_proc_read_table,
            per_proc_write=per_proc_write_table,
            per_file_read=per_file_read_table,
            per_file_write=per_file_write_table,
            per_proc_block_read=per_proc_block_read_table,
            per_proc_block_write=per_proc_block_write_table,
            per_disk_sector=per_disk_sector_table,
            per_disk_request=per_disk_request_table,
            per_disk_rtps=per_disk_rtps_table,
            per_netif_recv=per_netif_recv_table,
            per_netif_send=per_netif_send_table,
        )

    def _print_per_proc_io(self, result_table, title):
        header_format = '{:<25} {:<10} {:<10} {:<10}'
        label_header = header_format.format(
            'Process', 'Disk', 'Net', 'Unknown'
        )

        def get_label(row):
            label_format = '{:<25} {:>10} {:>10} {:>10}'
            if row.process.pid is None:
                pid_str = 'unknown (tid=%d)' % (row.process.tid)
            else:
                pid_str = str(row.process.pid)

            label = label_format.format(
                '%s (%s)' % (row.process.name, pid_str),
                format_utils.format_size(row.disk_size.value),
                format_utils.format_size(row.net_size.value),
                format_utils.format_size(row.unknown_size.value)
            )

            return label

        graph = termgraph.BarGraph(
            title='Per-process I/O ' + title,
            label_header=label_header,
            get_value=lambda row: row.size.value,
            get_value_str=format_utils.format_size,
            get_label=get_label,
            data=result_table.rows
        )

        graph.print_graph()

    def _print_per_proc_block_io(self, result_table, title):
        def get_label(row):
            proc_name = row.process.name

            if not proc_name:
                proc_name = 'unknown'

            if row.process.pid is None:
                pid_str = 'unknown (tid={})'.format(row.process.tid)
            else:
                pid_str = str(row.process.pid)

            return '{} (pid={})'.format(proc_name, pid_str)

        graph = termgraph.BarGraph(
            title='Block I/O ' + title,
            label_header='Process',
            get_value=lambda row: row.size.value,
            get_value_str=format_utils.format_size,
            get_label=get_label,
            data=result_table.rows
        )

        graph.print_graph()

    def _print_per_disk_sector(self, result_table):
        graph = termgraph.BarGraph(
            title='Disk Requests Sector Count',
            label_header='Disk',
            unit='sectors',
            get_value=lambda row: row.count.value,
            get_label=lambda row: row.disk.name,
            data=result_table.rows
        )

        graph.print_graph()

    def _print_per_disk_request(self, result_table):
        graph = termgraph.BarGraph(
            title='Disk Request Count',
            label_header='Disk',
            unit='requests',
            get_value=lambda row: row.count.value,
            get_label=lambda row: row.disk.name,
            data=result_table.rows
        )

        graph.print_graph()

    def _print_per_disk_rtps(self, result_table):
        graph = termgraph.BarGraph(
            title='Disk Request Average Latency',
            label_header='Disk',
            unit='ms',
            get_value=lambda row: row.rtps.value / 1000000,
            get_label=lambda row: row.disk.name,
            data=result_table.rows
        )

        graph.print_graph()

    def _print_per_netif_io(self, result_table, title):
        graph = termgraph.BarGraph(
            title='Network ' + title + ' Bytes',
            label_header='Interface',
            get_value=lambda row: row.size.value,
            get_value_str=format_utils.format_size,
            get_label=lambda row: row.netif.name,
            data=result_table.rows
        )

        graph.print_graph()

    def _print_per_file_io(self, result_table, title):
        # FIXME add option to show FD owners
        # FIXME why are read and write values the same?
        graph = termgraph.BarGraph(
            title='Per-file I/O ' + title,
            label_header='Path',
            get_value=lambda row: row.size.value,
            get_value_str=format_utils.format_size,
            get_label=lambda row: row.path.path,
            data=result_table.rows
        )

        graph.print_graph()

    def _print_usage(self, usage_tables):
        self._print_per_proc_io(usage_tables.per_proc_read, 'Read')
        self._print_per_proc_io(usage_tables.per_proc_write, 'Write')
        self._print_per_file_io(usage_tables.per_file_read, 'Read')
        self._print_per_file_io(usage_tables.per_file_write, 'Write')
        self._print_per_proc_block_io(usage_tables.per_proc_block_read, 'Read')
        self._print_per_proc_block_io(
            usage_tables.per_proc_block_write, 'Write'
        )
        self._print_per_disk_sector(usage_tables.per_disk_sector)
        self._print_per_disk_request(usage_tables.per_disk_request)
        self._print_per_disk_rtps(usage_tables.per_disk_rtps)
        self._print_per_netif_io(usage_tables.per_netif_recv, 'Received')
        self._print_per_netif_io(usage_tables.per_netif_send, 'Sent')

    def _fill_freq_result_table(self, duration_list, result_table):
        if not duration_list:
            return

        # The number of bins for the histogram
        resolution = self._args.freq_resolution

        min_duration = min(duration_list)
        max_duration = max(duration_list)
        # ns to µs
        min_duration /= 1000
        max_duration /= 1000

        step = (max_duration - min_duration) / resolution

        if step == 0:
            return

        buckets = []
        values = []

        for i in range(resolution):
            buckets.append(i * step)
            values.append(0)

        for duration in duration_list:
            duration /= 1000
            index = min(int((duration - min_duration) / step), resolution - 1)
            values[index] += 1

        for index, value in enumerate(values):
            result_table.append_row(
                latency_lower=mi.Duration.from_us(index * step + min_duration),
                latency_upper=mi.Duration.from_us((index + 1) * step +
                                                  min_duration),
                count=mi.Number(value),
            )

    def _get_disk_freq_result_tables(self, period_data, begin, end):
        result_tables = []

        for disk in period_data.disks.values():
            rq_durations = [rq.duration for rq in disk.rq_list if
                            self._filter_io_request(rq)]
            subtitle = 'disk: {}'.format(disk.diskname)
            result_table = \
                self._mi_create_result_table(self._MI_TABLE_CLASS_FREQ,
                                             begin, end, subtitle)
            self._fill_freq_result_table(rq_durations, result_table)
            result_tables.append(result_table)

        return result_tables

    def _get_syscall_freq_result_tables(self, period_data, begin, end):
        open_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_FREQ,
                                         begin, end, 'open')
        read_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_FREQ,
                                         begin, end, 'read')
        write_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_FREQ,
                                         begin, end, 'write')
        sync_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_FREQ,
                                         begin, end, 'sync')
        self._fill_freq_result_table(
            [io_rq.duration for io_rq in
             self._analysis.open_io_requests(period_data)
             if self._filter_io_request(io_rq)], open_table)
        self._fill_freq_result_table(
            [io_rq.duration for io_rq in
             self._analysis.read_io_requests(period_data)
             if self._filter_io_request(io_rq)], read_table)
        self._fill_freq_result_table(
            [io_rq.duration for io_rq in
             self._analysis.write_io_requests(period_data)
             if self._filter_io_request(io_rq)], write_table)
        self._fill_freq_result_table(
            [io_rq.duration for io_rq in
             self._analysis.sync_io_requests(period_data)
             if self._filter_io_request(io_rq)], sync_table)

        return [open_table, read_table, write_table, sync_table]

    def _get_freq_result_tables(self, period_data, begin, end):
        syscall_tables = self._get_syscall_freq_result_tables(period_data,
                                                              begin, end)
        disk_tables = self._get_disk_freq_result_tables(period_data,
                                                        begin, end)

        return syscall_tables + disk_tables

    def _print_one_freq(self, result_table):
        graph = termgraph.FreqGraph(
            data=result_table.rows,
            get_value=lambda row: row.count.value,
            get_lower_bound=lambda row: row.latency_lower.to_us(),
            title='{} {}'.format(result_table.title, result_table.subtitle),
            unit='µs'
        )

        graph.print_graph()

    def _print_freq(self, freq_tables):
        for freq_table in freq_tables:
            self._print_one_freq(freq_table)

    def _append_log_row(self, period_data, io_rq, result_table):
        if io_rq.size is None:
            size = mi.Empty()
        else:
            size = mi.Size(io_rq.size)

        tid = io_rq.tid
        proc_stats = period_data.tids[tid]
        proc_name = proc_stats.comm

        # TODO: handle fd_in/fd_out for RW type operations
        if io_rq.fd is None:
            path = mi.Empty()
            fd = mi.Empty()
        else:
            fd = mi.Fd(io_rq.fd)
            parent_proc = proc_stats

            if parent_proc.pid is not None:
                parent_proc = period_data.tids[parent_proc.pid]

            fd_stats = parent_proc.get_fd(io_rq.fd, io_rq.end_ts)

            if fd_stats is not None:
                path = mi.Path(fd_stats.filename)
            else:
                path = mi.Unknown()

        result_table.append_row(
            time_range=mi.TimeRange(io_rq.begin_ts, io_rq.end_ts),
            out_of_range=mi.Boolean(self._is_io_rq_out_of_range(io_rq)),
            duration=mi.Duration(io_rq.duration),
            syscall=mi.Syscall(io_rq.syscall_name),
            size=size,
            process=mi.Process(proc_name, tid=tid),
            path=path,
            fd=fd,
        )

    def _fill_log_result_table(self, period_data, rq_list, sort_key, is_top,
                               result_table):
        if not rq_list:
            return

        count = 0

        for io_rq in sorted(rq_list, key=operator.attrgetter(sort_key),
                            reverse=is_top):
            if is_top and count > self._args.limit:
                break

            self._append_log_row(period_data, io_rq, result_table)
            count += 1

    def _fill_log_result_table_from_io_requests(self, period_data, io_requests,
                                                sort_key, is_top,
                                                result_table):
        io_requests = [io_rq for io_rq in io_requests if
                       self._filter_io_request(io_rq)]
        self._fill_log_result_table(period_data, io_requests, sort_key, is_top,
                                    result_table)

    def _get_top_result_tables(self, period_data, begin, end):
        open_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_TOP_SYSCALL,
                                         begin, end, 'open')
        read_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_TOP_SYSCALL,
                                         begin, end, 'read')
        write_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_TOP_SYSCALL,
                                         begin, end, 'write')
        sync_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_TOP_SYSCALL,
                                         begin, end, 'sync')
        self._fill_log_result_table_from_io_requests(
            period_data, self._analysis.open_io_requests(period_data),
            'duration', True, open_table)
        self._fill_log_result_table_from_io_requests(
            period_data, self._analysis.read_io_requests(period_data),
            'duration', True, read_table)
        self._fill_log_result_table_from_io_requests(
            period_data, self._analysis.write_io_requests(period_data),
            'duration', True, write_table)
        self._fill_log_result_table_from_io_requests(
            period_data, self._analysis.sync_io_requests(period_data),
            'duration', True, sync_table)

        return [open_table, read_table, write_table, sync_table]

    def _print_log_row(self, row):
        fmt = '{:<40} {:<16} {:>16} {:>11}  {:<24} {:<8} {:<14}'
        time_range_str = format_utils.format_time_range(
            row.time_range.begin.value,
            row.time_range.end.value,
            self._args.multi_day,
            self._args.gmt
        )
        duration_str = '%0.03f' % row.duration.to_us()

        if type(row.size) is mi.Empty:
            size = 'N/A'
        else:
            size = format_utils.format_size(row.size.value)

        tid = row.process.tid
        proc_name = row.process.name

        if type(row.fd) is mi.Empty:
            file_str = 'N/A'
        else:
            if type(row.path) is mi.Unknown:
                path = 'unknown'
            else:
                path = row.path.path

            file_str = '%s (fd=%s)' % (path, row.fd.fd)

        if row.out_of_range.value:
            time_range_str += '*'
            duration_str += '*'
        else:
            time_range_str += ' '
            duration_str += ' '

        print(fmt.format(time_range_str, row.syscall.name, duration_str,
                         size, proc_name, tid, file_str))

    def _print_log(self, result_table):
        if not result_table.rows:
            return

        has_out_of_range_rq = False

        print()
        fmt = '{} {} (usec)'
        print(fmt.format(result_table.title, result_table.subtitle))
        header_fmt = '{:<20} {:<20} {:<16} {:<23} {:<5} {:<24} {:<8} {:<14}'
        print(header_fmt.format(
            'Begin', 'End', 'Name', 'Duration (usec)', 'Size', 'Proc', 'PID',
            'Filename'))

        for row in result_table.rows:
            self._print_log_row(row)

            if not has_out_of_range_rq and row.out_of_range.value:
                has_out_of_range_rq = True

        if has_out_of_range_rq:
            print('*: Syscalls started and/or completed outside of the '
                  'range specified')

    def _print_top(self, top_tables):
        for table in top_tables:
            self._print_log(table)

    def _get_log_result_table(self, period_data, begin, end):
        log_table = self._mi_create_result_table(self._MI_TABLE_CLASS_LOG,
                                                 begin, end)
        self._fill_log_result_table_from_io_requests(
            period_data, self._analysis.io_requests(period_data),
            'begin_ts', False, log_table)

        return log_table

    def _append_latency_stats_row(self, obj, rq_durations, result_table):
        rq_count = len(rq_durations)
        total_duration = sum(rq_durations)

        if len(rq_durations) > 0:
            min_duration = min(rq_durations)
            max_duration = max(rq_durations)
        else:
            min_duration = 0
            max_duration = 0

        if rq_count < 2:
            stdev = mi.Unknown()
        else:
            stdev = mi.Duration(statistics.stdev(rq_durations))

        if rq_count > 0:
            avg = total_duration / rq_count
        else:
            avg = 0

        result_table.append_row(
            obj=obj,
            count=mi.Number(rq_count),
            min_latency=mi.Duration(min_duration),
            avg_latency=mi.Duration(avg),
            max_latency=mi.Duration(max_duration),
            stdev_latency=stdev,
        )

    def _append_latency_stats_row_from_requests(self, obj, io_requests,
                                                result_table):
        rq_durations = [io_rq.duration for io_rq in io_requests if
                        self._filter_io_request(io_rq)]
        self._append_latency_stats_row(obj, rq_durations, result_table)

    def _get_syscall_latency_stats_result_table(self, period_data, begin, end):
        result_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_SYSCALL_LATENCY_STATS, begin, end)
        append_fn = self._append_latency_stats_row_from_requests
        append_fn(mi.String('Open'),
                  self._analysis.open_io_requests(period_data), result_table)
        append_fn(mi.String('Read'),
                  self._analysis.read_io_requests(period_data), result_table)
        append_fn(mi.String('Write'),
                  self._analysis.write_io_requests(period_data), result_table)
        append_fn(mi.String('Sync'),
                  self._analysis.sync_io_requests(period_data), result_table)

        return result_table

    def _get_disk_latency_stats_result_table(self, period_data, begin, end):
        if not period_data.disks:
            return

        result_table = self._mi_create_result_table(
            self._MI_TABLE_CLASS_PART_LATENCY_STATS, begin, end)

        for disk in period_data.disks.values():
            if disk.rq_count:
                rq_durations = [rq.duration for rq in disk.rq_list if
                                self._filter_io_request(rq)]
                disk = mi.Disk(disk.diskname)
                self._append_latency_stats_row(disk, rq_durations,
                                               result_table)

        return result_table

    def _get_latency_stats_result_tables(self, period_data, begin, end):
        syscall_tbl = self._get_syscall_latency_stats_result_table(period_data,
                                                                   begin, end)
        disk_tbl = self._get_disk_latency_stats_result_table(period_data,
                                                             begin, end)

        return syscall_tbl, disk_tbl

    def _print_latency_stats_row(self, row):
        if type(row.stdev_latency) is mi.Unknown:
            stdev = '?'
        else:
            stdev = '%0.03f' % row.stdev_latency.to_us()

        avg = '%0.03f' % row.avg_latency.to_us()
        min_duration = '%0.03f' % row.min_latency.to_us()
        max_duration = '%0.03f' % row.max_latency.to_us()

        print(IoAnalysisCommand._LATENCY_STATS_FORMAT.format(
            str(row.obj), row.count.value, min_duration,
            avg, max_duration, stdev))

    def _print_syscall_latency_stats(self, stats_table):
        print('\nSyscalls latency statistics (usec):')
        print(IoAnalysisCommand._LATENCY_STATS_FORMAT.format(
            'Type', 'Count', 'Min', 'Average', 'Max', 'Stdev'))
        print(IoAnalysisCommand._SECTION_SEPARATOR_STRING)

        for row in stats_table.rows:
            self._print_latency_stats_row(row)

    def _print_disk_latency_stats(self, stats_table):
        if not stats_table or not stats_table.rows:
            return

        print('\nDisk latency statistics (usec):')
        print(IoAnalysisCommand._LATENCY_STATS_FORMAT.format(
            'Name', 'Count', 'Min', 'Average', 'Max', 'Stdev'))
        print(IoAnalysisCommand._SECTION_SEPARATOR_STRING)

        for row in stats_table.rows:
            self._print_latency_stats_row(row)

    def _print_latency_stats(self, syscall_latency_stats_table,
                             disk_latency_stats_table):
        self._print_syscall_latency_stats(syscall_latency_stats_table)
        self._print_disk_latency_stats(disk_latency_stats_table)

    def _add_arguments(self, ap):
        Command._add_min_max_args(ap)
        Command._add_log_args(
            ap, help='Output the I/O requests in chronological order')
        Command._add_top_args(
            ap, help='Output the top I/O latencies by category')
        Command._add_stats_args(ap, help='Output the I/O latency statistics')
        Command._add_freq_args(
            ap, help='Output the I/O latency frequency distribution')
        ap.add_argument('--usage', action='store_true',
                        help='Output the I/O usage')
        ap.add_argument('--minsize', type=float,
                        help='Filter out, I/O operations working with '
                        'less that minsize bytes')
        ap.add_argument('--maxsize', type=float,
                        help='Filter out, I/O operations working with '
                        'more that maxsize bytes')


def _run(mi_mode):
    iocmd = IoAnalysisCommand(mi_mode=mi_mode)
    iocmd.run()


def _runstats(mi_mode):
    sys.argv.insert(1, '--stats')
    _run(mi_mode)


def _runlog(mi_mode):
    sys.argv.insert(1, '--log')
    _run(mi_mode)


def _runfreq(mi_mode):
    sys.argv.insert(1, '--freq')
    _run(mi_mode)


def _runlatencytop(mi_mode):
    sys.argv.insert(1, '--top')
    _run(mi_mode)


def _runusage(mi_mode):
    sys.argv.insert(1, '--usage')
    _run(mi_mode)


def runstats():
    _runstats(mi_mode=False)


def runlog():
    _runlog(mi_mode=False)


def runfreq():
    _runfreq(mi_mode=False)


def runlatencytop():
    _runlatencytop(mi_mode=False)


def runusage():
    _runusage(mi_mode=False)


def runstats_mi():
    _runstats(mi_mode=True)


def runlog_mi():
    _runlog(mi_mode=True)


def runfreq_mi():
    _runfreq(mi_mode=True)


def runlatencytop_mi():
    _runlatencytop(mi_mode=True)


def runusage_mi():
    _runusage(mi_mode=True)



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from . import sp, sv


class SyscallsStateProvider(sp.StateProvider):
    def __init__(self, state):
        cbs = {
            'syscall_entry': self._process_syscall_entry,
            'syscall_exit': self._process_syscall_exit
        }

        super().__init__(state, cbs)

    def _process_syscall_entry(self, event):
        cpu_id = event['cpu_id']

        if cpu_id not in self._state.cpus:
            return

        cpu = self._state.cpus[cpu_id]
        if cpu.current_tid is None:
            return

        proc = self._state.tids[cpu.current_tid]
        proc.current_syscall = sv.SyscallEvent.new_from_entry(event)

    def _process_syscall_exit(self, event):
        cpu_id = event['cpu_id']
        if cpu_id not in self._state.cpus:
            return

        cpu = self._state.cpus[cpu_id]
        if cpu.current_tid is None:
            return

        proc = self._state.tids[cpu.current_tid]
        current_syscall = proc.current_syscall
        if current_syscall is None:
            return

        current_syscall.process_exit(event)

        self._state.send_notification_cb('syscall_exit',
                                         proc=proc,
                                         event=event,
                                         cpu_id=cpu_id)

        # If it's an IO Syscall, the IO state provider will take care of
        # clearing the current syscall, so only clear here if it's not
        if current_syscall.name not in sv.SyscallConsts.IO_SYSCALLS:
            self._state.tids[cpu.current_tid].current_syscall = None



# The MIT License (MIT)
#
# Copyright (C) 2016 - Julien Desfossez <jdesfossez@efficios.com>
#                      Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import subprocess
import unittest
import locale
from .trace_writer import TraceWriter


class AnalysisTest(unittest.TestCase):
    COMMON_OPTIONS = '--no-color --no-progress --skip-validation --gmt'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.rm_trace = True

    def set_up_class(self):
        dirname = os.path.dirname(os.path.realpath(__file__))
        self.data_path = dirname + '/expected/'
        self.maxDiff = None
        self.trace_writer = TraceWriter()
        self.write_trace()

    def tear_down_class(self):
        if self.rm_trace:
            self.trace_writer.rm_trace()

    def write_trace(self):
        raise NotImplementedError

    def run(self, result=None):
        self.set_up_class()
        super().run(result)
        self.tear_down_class()

        return result

    def get_expected_output(self, test_name):
        expected_path = os.path.join(self.data_path, test_name + '.txt')
        with open(expected_path, 'r', encoding='utf-8') as expected_file:
            return expected_file.read()

    def _test_locale(self, locale_name):
        try:
            locale.setlocale(locale.LC_ALL, locale_name)
            return True
        except locale.Error:
            return False

    def _get_utf8_locale(self):
        # Test the two most common UTF-8 locales
        if self._test_locale('C.UTF-8'):
            return 'C.UTF-8'
        if self._test_locale('en_US.UTF-8'):
            return 'en_US.UTF-8'
        print('No supported UTF-8 locale detected')
        raise NameError

    def get_cmd_output(self, exec_name, options=''):
        cmd_fmt = './{} {} {} {}'
        cmd = cmd_fmt.format(exec_name, self.COMMON_OPTIONS,
                             options, self.trace_writer.trace_root)

        # Create an utf-8 test env
        test_locale = self._get_utf8_locale()
        test_env = os.environ.copy()
        test_env['LC_ALL'] = test_locale

        process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT, env=test_env)
        output, unused_err = process.communicate()
        output = output.decode('utf-8')

        if output[-1:] == '\n':
            output = output[:-1]

        return output

    def save_test_result(self, result, test_name):
        result_path = os.path.join(self.trace_writer.trace_root, test_name)
        with open(result_path, 'w', encoding='utf-8') as result_file:
            result_file.write(result)
            self.rm_trace = False

    def _assertMultiLineEqual(self, result, expected, test_name):
        try:
            self.assertMultiLineEqual(result, expected)
        except AssertionError:
            self.save_test_result(result, test_name)
            raise



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#               2015 - Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import operator
from ..common import format_utils
from .command import Command
from ..core import cputop
from . import mi
from . import termgraph


class Cputop(Command):
    _DESC = """The cputop command."""
    _ANALYSIS_CLASS = cputop.Cputop
    _MI_TITLE = 'Top CPU usage'
    _MI_DESCRIPTION = 'Per-TID, per-CPU, and total top CPU usage'
    _MI_TAGS = [mi.Tags.CPU, mi.Tags.TOP]
    _MI_TABLE_CLASS_PER_PROC = 'per-process'
    _MI_TABLE_CLASS_PER_CPU = 'per-cpu'
    _MI_TABLE_CLASS_TOTAL = 'total'
    _MI_TABLE_CLASS_SUMMARY = 'summary'
    _MI_TABLE_CLASSES = [
        (
            _MI_TABLE_CLASS_PER_PROC,
            'Per-TID top CPU usage', [
                ('process', 'Process', mi.Process),
                ('migrations', 'Migration count', mi.Number, 'migrations'),
                ('prio_list', 'Chronological priorities', mi.String),
                ('usage', 'CPU usage', mi.Ratio),
            ]
        ),
        (
            _MI_TABLE_CLASS_PER_CPU,
            'Per-CPU top CPU usage', [
                ('cpu', 'CPU', mi.Cpu),
                ('usage', 'CPU usage', mi.Ratio),
            ]),
        (
            _MI_TABLE_CLASS_TOTAL,
            'Total CPU usage', [
                ('usage', 'CPU usage', mi.Ratio),
            ]
        ),
        (
            _MI_TABLE_CLASS_SUMMARY,
            'CPU usage - summary', [
                ('time_range', 'Time range', mi.TimeRange),
                ('usage', 'Total CPU usage', mi.Ratio),
            ]
        ),
    ]

    def _analysis_tick(self, period_data, end_ns):
        if period_data is None:
            return

        begin_ns = period_data.period.begin_evt.timestamp
        per_tid_table = self._get_per_tid_usage_result_table(period_data,
                                                             begin_ns, end_ns)
        per_cpu_table = self._get_per_cpu_usage_result_table(period_data,
                                                             begin_ns, end_ns)
        total_table = self._get_total_usage_result_table(period_data, begin_ns,
                                                         end_ns)

        if self._mi_mode:
            self._mi_append_result_table(per_tid_table)
            self._mi_append_result_table(per_cpu_table)
            self._mi_append_result_table(total_table)
        else:
            self._print_date(begin_ns, end_ns)
            self._print_per_tid_usage(per_tid_table)
            self._print_per_cpu_usage(per_cpu_table)

            if total_table:
                self._print_total_cpu_usage(total_table)

    def _create_summary_result_tables(self):
        total_tables = self._mi_get_result_tables(self._MI_TABLE_CLASS_TOTAL)
        begin = total_tables[0].timerange.begin.value
        end = total_tables[-1].timerange.end.value
        summary_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_SUMMARY,
                                         begin, end)

        for total_table in total_tables:
            usage = total_table.rows[0].usage
            summary_table.append_row(
                time_range=total_table.timerange,
                usage=usage,
            )

        self._mi_clear_result_tables()
        self._mi_append_result_table(summary_table)

    def _get_per_tid_usage_result_table(self, period_data, begin_ns, end_ns):
        result_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_PER_PROC,
                                         begin_ns, end_ns)
        count = 0

        for tid in sorted(period_data.tids.values(),
                          key=operator.attrgetter('usage_percent'),
                          reverse=True):
            prio_list = format_utils.format_prio_list(tid.prio_list)

            result_table.append_row(
                process=mi.Process(tid.comm, tid=tid.tid),
                migrations=mi.Number(tid.migrate_count),
                prio_list=mi.String(prio_list),
                usage=mi.Ratio.from_percentage(tid.usage_percent)
            )
            count += 1

            if self._args.limit > 0 and count >= self._args.limit:
                break

        return result_table

    def _get_per_cpu_usage_result_table(self, period_data, begin_ns, end_ns):
        result_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_PER_CPU,
                                         begin_ns, end_ns)

        for cpu in sorted(period_data.cpus.values(),
                          key=operator.attrgetter('cpu_id')):
            result_table.append_row(
                cpu=mi.Cpu(cpu.cpu_id),
                usage=mi.Ratio.from_percentage(cpu.usage_percent)
            )

        return result_table

    def _get_total_usage_result_table(self, period_data, begin_ns, end_ns):
        result_table = \
            self._mi_create_result_table(self._MI_TABLE_CLASS_TOTAL,
                                         begin_ns, end_ns)

        cpu_count = len(self.state.cpus)
        usage_percent = 0

        if not cpu_count:
            return

        for cpu in sorted(period_data.cpus.values(),
                          key=operator.attrgetter('usage_percent'),
                          reverse=True):
            usage_percent += cpu.usage_percent

        # average per CPU
        usage_percent /= cpu_count
        result_table.append_row(
            usage=mi.Ratio.from_percentage(usage_percent),
        )

        return result_table

    def _print_per_tid_usage(self, result_table):
        row_format = '  {:<25} {:>10}   {}'
        label_header = row_format.format('Process', 'Migrations', 'Priorities')

        def format_label(row):
            return row_format.format(
                '%s (%d)' % (row.process.name, row.process.tid),
                row.migrations.value,
                row.prio_list.value,
            )

        graph = termgraph.BarGraph(
            title='Per-TID Usage',
            unit='%',
            get_value=lambda row: row.usage.to_percentage(),
            get_label=format_label,
            label_header=label_header,
            data=result_table.rows
        )

        graph.print_graph()

    def _print_per_cpu_usage(self, result_table):
        graph = termgraph.BarGraph(
            title='Per-CPU Usage',
            unit='%',
            get_value=lambda row: row.usage.to_percentage(),
            get_label=lambda row: 'CPU %d' % row.cpu.id,
            data=result_table.rows
        )

        graph.print_graph()

    def _print_total_cpu_usage(self, result_table):
        usage_percent = result_table.rows[0].usage.to_percentage()
        print('\nTotal CPU Usage: %0.02f%%\n' % usage_percent)

    def _add_arguments(self, ap):
        Command._add_proc_filter_args(ap)
        Command._add_top_args(ap)


def _run(mi_mode):
    cputopcmd = Cputop(mi_mode=mi_mode)
    cputopcmd.run()


def run():
    _run(mi_mode=False)


def run_mi():
    _run(mi_mode=True)



# The MIT License (MIT)
#
# Copyright (C) 2015 - Julien Desfossez <jdesfossez@efficios.com>
#               2015 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import socket
from ..common import format_utils, trace_utils


class Process():
    def __init__(self, tid=None, pid=None, comm='', prio=None):
        self.tid = tid
        self.pid = pid
        self.comm = comm
        self.prio = prio
        # indexed by fd
        self.fds = {}
        self.current_syscall = None
        # the process scheduled before this one
        self.prev_tid = None
        self.last_wakeup = None
        self.last_waker = None


class CPU():
    def __init__(self, cpu_id):
        self.cpu_id = cpu_id
        self.current_tid = None
        self.current_hard_irq = None
        # softirqs use a dict because multiple ones can be raised before
        # handling. They are indexed by vec, and each entry is a list,
        # ordered chronologically
        self.current_softirqs = {}


class MemoryManagement():
    def __init__(self):
        self.page_count = 0


class SyscallEvent():
    def __init__(self, name, begin_ts):
        self.name = name
        self.begin_ts = begin_ts
        self.end_ts = None
        self.ret = None
        self.duration = None
        # Only applicable to I/O syscalls
        self.io_rq = None

    def process_exit(self, event):
        self.end_ts = event.timestamp
        # On certain architectures (notably arm32), lttng-modules
        # versions prior to 2.8 would erroneously trace certain
        # syscalls (e.g. mmap2) without their return value. In this
        # case, get() will simply set self.ret to None. These syscalls
        # with a None return value should simply be ignored down the
        # line.
        self.ret = event.get('ret')
        self.duration = self.end_ts - self.begin_ts

    @classmethod
    def new_from_entry(cls, event):
        name = trace_utils.get_syscall_name(event)
        return cls(name, event.timestamp)


class Disk():
    def __init__(self, dev, diskname=None):
        self.dev = dev
        self.diskname = diskname
        # pending block IO Requests, indexed by sector
        self.pending_requests = {}


class FDType():
    unknown = 0
    disk = 1
    net = 2
    # not 100% sure they are network FDs (assumed when net_dev_xmit is
    # called during a write syscall and the type in unknown).
    maybe_net = 3

    @staticmethod
    def get_fd_type(name, family):
        if name in SyscallConsts.NET_OPEN_SYSCALLS:
            if family in SyscallConsts.INET_FAMILIES:
                return FDType.net
            if family in SyscallConsts.DISK_FAMILIES:
                return FDType.disk

        if name in SyscallConsts.DISK_OPEN_SYSCALLS:
            return FDType.disk

        return FDType.unknown


class FD():
    def __init__(self, fd, filename='unknown', fd_type=FDType.unknown,
                 cloexec=False, family=None):
        self.fd = fd
        self.filename = filename
        self.fd_type = fd_type
        self.cloexec = cloexec
        self.family = family

    @classmethod
    def new_from_fd(cls, fd):
        return cls(fd.fd, fd.filename, fd.fd_type, fd.cloexec, fd.family)

    @classmethod
    def new_from_open_rq(cls, io_rq):
        return cls(io_rq.fd, io_rq.filename, io_rq.fd_type, io_rq.cloexec,
                   io_rq.family)


class IRQ():
    def __init__(self, id, cpu_id, begin_ts=None):
        self.id = id
        self.cpu_id = cpu_id
        self.begin_ts = begin_ts
        self.end_ts = None

    @property
    def duration(self):
        if not self.end_ts or not self.begin_ts:
            return None

        return self.end_ts - self.begin_ts


class HardIRQ(IRQ):
    def __init__(self, id, cpu_id, begin_ts):
        super().__init__(id, cpu_id, begin_ts)
        self.ret = None

    @classmethod
    def new_from_irq_handler_entry(cls, event):
        id = event['irq']
        cpu_id = event['cpu_id']
        begin_ts = event.timestamp
        return cls(id, cpu_id, begin_ts)


class SoftIRQ(IRQ):
    def __init__(self, id, cpu_id, raise_ts=None, begin_ts=None):
        super().__init__(id, cpu_id, begin_ts)
        self.raise_ts = raise_ts

    @classmethod
    def new_from_softirq_raise(cls, event):
        id = event['vec']
        cpu_id = event['cpu_id']
        raise_ts = event.timestamp
        return cls(id, cpu_id, raise_ts)

    @classmethod
    def new_from_softirq_entry(cls, event):
        id = event['vec']
        cpu_id = event['cpu_id']
        begin_ts = event.timestamp
        return cls(id, cpu_id, begin_ts=begin_ts)


class IORequest():
    # I/O operations
    OP_OPEN = 1
    OP_READ = 2
    OP_WRITE = 3
    OP_CLOSE = 4
    OP_SYNC = 5
    # Operation used for requests that both read and write,
    # e.g. splice and sendfile
    OP_READ_WRITE = 6

    def __init__(self, begin_ts, size, tid, operation):
        self.begin_ts = begin_ts
        self.end_ts = None
        self.duration = None
        # request size in bytes
        self.size = size
        self.operation = operation
        # tid of process that triggered the rq
        self.tid = tid
        # Error number if request failed
        self.errno = None

    @staticmethod
    def is_equivalent_operation(left_op, right_op):
        """Predicate used to compare equivalence of IO_OPERATION.

        This method is employed because OP_READ_WRITE behaves like a
        set containing both OP_READ and OP_WRITE and is therefore
        equivalent to these operations as well as itself
        """
        if left_op == IORequest.OP_READ_WRITE:
            return right_op in [IORequest.OP_READ, IORequest.OP_WRITE,
                                IORequest.OP_READ_WRITE]
        if left_op == IORequest.OP_READ:
            return right_op in [IORequest.OP_READ, IORequest.OP_READ_WRITE]
        if left_op == IORequest.OP_WRITE:
            return right_op in [IORequest.OP_WRITE, IORequest.OP_READ_WRITE]

        return left_op == right_op


class SyscallIORequest(IORequest):
    def __init__(self, begin_ts, size, tid, operation, syscall_name):
        super().__init__(begin_ts, None, tid, operation)
        self.fd = None
        self.syscall_name = syscall_name
        # Number of pages alloc'd/freed/written to disk during the rq
        self.pages_allocated = 0
        self.pages_freed = 0
        self.pages_written = 0
        # Whether kswapd was forced to wakeup during the rq
        self.woke_kswapd = False

    def update_from_exit(self, event):
        self.end_ts = event.timestamp
        self.duration = self.end_ts - self.begin_ts
        if event['ret'] < 0:
            self.errno = -event['ret']


class OpenIORequest(SyscallIORequest):
    def __init__(self, begin_ts, tid, syscall_name, filename,
                 fd_type):
        super().__init__(begin_ts, None, tid, IORequest.OP_OPEN, syscall_name)
        # FD set on syscall exit
        self.fd = None
        self.filename = filename
        self.fd_type = fd_type
        self.family = None
        self.cloexec = False

    def update_from_exit(self, event):
        super().update_from_exit(event)
        if event['ret'] >= 0:
            self.fd = event['ret']

    @classmethod
    def new_from_disk_open(cls, event, tid):
        begin_ts = event.timestamp
        name = trace_utils.get_syscall_name(event)
        filename = event['filename']

        req = cls(begin_ts, tid, name, filename, FDType.disk)
        req.cloexec = event['flags'] & os.O_CLOEXEC == os.O_CLOEXEC

        return req

    @classmethod
    def new_from_accept(cls, event, tid):
        # Handle both accept and accept4
        begin_ts = event.timestamp
        name = trace_utils.get_syscall_name(event)
        req = cls(begin_ts, tid, name, 'socket', FDType.net)

        if 'family' in event:
            req.family = event['family']
            # Set filename to ip:port if INET socket
            if req.family == socket.AF_INET:
                req.filename = format_utils.format_ipv4(
                    event['v4addr'], event['sport']
                )

        return req

    @classmethod
    def new_from_socket(cls, event, tid):
        begin_ts = event.timestamp
        req = cls(begin_ts, tid, 'socket', 'socket', FDType.net)

        if 'family' in event:
            req.family = event['family']

        return req

    @classmethod
    def new_from_old_fd(cls, event, tid, old_fd):
        begin_ts = event.timestamp
        name = trace_utils.get_syscall_name(event)
        if old_fd is None:
            filename = 'unknown'
            fd_type = FDType.unknown
        else:
            filename = old_fd.filename
            fd_type = old_fd.fd_type

        return cls(begin_ts, tid, name, filename, fd_type)


class CloseIORequest(SyscallIORequest):
    def __init__(self, begin_ts, tid, fd):
        super().__init__(begin_ts, None, tid, IORequest.OP_CLOSE, 'close')
        self.fd = fd


class ReadWriteIORequest(SyscallIORequest):
    def __init__(self, begin_ts, size, tid, operation, syscall_name):
        super().__init__(begin_ts, size, tid, operation, syscall_name)
        # The size returned on syscall exit, in bytes. May differ from
        # the size initially requested
        self.returned_size = None
        # Unused if fd is set
        self.fd_in = None
        self.fd_out = None

    def update_from_exit(self, event):
        super().update_from_exit(event)
        ret = event['ret']
        if ret >= 0:
            self.returned_size = ret
            # Set the size to the returned one if none was set at
            # entry, as with recvmsg or sendmsg
            if self.size is None:
                self.size = ret

    @classmethod
    def new_from_splice(cls, event, tid):
        begin_ts = event.timestamp
        size = event['len']

        req = cls(begin_ts, size, tid, IORequest.OP_READ_WRITE, 'splice')
        req.fd_in = event['fd_in']
        req.fd_out = event['fd_out']

        return req

    @classmethod
    def new_from_sendfile64(cls, event, tid):
        begin_ts = event.timestamp
        size = event['count']

        req = cls(begin_ts, size, tid, IORequest.OP_READ_WRITE, 'sendfile64')
        req.fd_in = event['in_fd']
        req.fd_out = event['out_fd']

        return req

    @classmethod
    def new_from_fd_event(cls, event, tid, size_key):
        begin_ts = event.timestamp
        # Some events, like recvmsg or sendmsg, only have size info on return
        if size_key is not None:
            size = event[size_key]
        else:
            size = None

        syscall_name = trace_utils.get_syscall_name(event)
        if syscall_name in SyscallConsts.READ_SYSCALLS:
            operation = IORequest.OP_READ
        else:
            operation = IORequest.OP_WRITE

        req = cls(begin_ts, size, tid, operation, syscall_name)
        req.fd = event['fd']

        return req


class SyncIORequest(SyscallIORequest):
    def __init__(self, begin_ts, size, tid, syscall_name):
        super().__init__(begin_ts, size, tid, IORequest.OP_SYNC, syscall_name)

    @classmethod
    def new_from_sync(cls, event, tid):
        begin_ts = event.timestamp
        size = None

        return cls(begin_ts, size, tid, 'sync')

    @classmethod
    def new_from_fsync(cls, event, tid):
        # Also handle fdatasync
        begin_ts = event.timestamp
        size = None
        syscall_name = trace_utils.get_syscall_name(event)

        req = cls(begin_ts, size, tid, syscall_name)
        req.fd = event['fd']

        return req

    @classmethod
    def new_from_sync_file_range(cls, event, tid):
        begin_ts = event.timestamp
        size = event['nbytes']

        req = cls(begin_ts, size, tid, 'sync_file_range')
        req.fd = event['fd']

        return req


class BlockIORequest(IORequest):
    # Logical sector size in bytes, according to the kernel
    SECTOR_SIZE = 512

    def __init__(self, begin_ts, tid, operation, dev, sector, nr_sector):
        size = nr_sector * BlockIORequest.SECTOR_SIZE
        super().__init__(begin_ts, size, tid, operation)
        self.dev = dev
        self.sector = sector
        self.nr_sector = nr_sector

    def update_from_rq_complete(self, event):
        self.end_ts = event.timestamp
        self.duration = self.end_ts - self.begin_ts

    @classmethod
    def new_from_rq_issue(cls, event):
        begin_ts = event.timestamp
        dev = event['dev']
        sector = event['sector']
        nr_sector = event['nr_sector']
        tid = event['tid']
        # An even rwbs indicates read operation, odd indicates write
        if event['rwbs'] % 2 == 0:
            operation = IORequest.OP_READ
        else:
            operation = IORequest.OP_WRITE

        return cls(begin_ts, tid, operation, dev, sector, nr_sector)


class BlockRemapRequest():
    def __init__(self, dev, sector, old_dev, old_sector):
        self.dev = dev
        self.sector = sector
        self.old_dev = old_dev
        self.old_sector = old_sector


class SyscallConsts():
    # TODO: decouple socket/family logic from this class
    INET_FAMILIES = [socket.AF_INET, socket.AF_INET6]
    DISK_FAMILIES = [socket.AF_UNIX]
    # list nof syscalls that open a FD on disk (in the exit_syscall event)
    DISK_OPEN_SYSCALLS = ['open', 'openat']
    # list of syscalls that open a FD on the network
    # (in the exit_syscall event)
    NET_OPEN_SYSCALLS = ['socket']
    # list of syscalls that can duplicate a FD
    DUP_OPEN_SYSCALLS = ['fcntl', 'dup', 'dup2', 'dup3']
    SYNC_SYSCALLS = ['sync', 'sync_file_range', 'fsync', 'fdatasync']
    # merge the 3 open lists
    OPEN_SYSCALLS = DISK_OPEN_SYSCALLS + NET_OPEN_SYSCALLS + DUP_OPEN_SYSCALLS
    # list of syscalls that close a FD (in the 'fd =' field)
    CLOSE_SYSCALLS = ['close']
    # list of syscall that read on a FD, value in the exit_syscall following
    READ_SYSCALLS = ['read', 'recvmsg', 'recvfrom', 'readv', 'pread',
                     'pread64', 'preadv']
    # list of syscall that write on a FD, value in the exit_syscall following
    WRITE_SYSCALLS = ['write', 'sendmsg', 'sendto', 'writev', 'pwrite',
                      'pwrite64', 'pwritev']
    # list of syscalls that both read and write on two FDs
    READ_WRITE_SYSCALLS = ['splice', 'sendfile64']
    # All I/O related syscalls
    IO_SYSCALLS = OPEN_SYSCALLS + CLOSE_SYSCALLS + READ_SYSCALLS + \
        WRITE_SYSCALLS + SYNC_SYSCALLS + READ_WRITE_SYSCALLS



# The MIT License (MIT)
#
# Copyright (C) 2016 - Philippe Proulx <pproulx@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pyparsing as pp
from ..core import period


class MalformedExpression(Exception):
    pass


class DuplicatePeriodCapture(Exception):
    def __init__(self, name):
        self._name = name

    def __str__(self):
        return 'Duplicate period capture name: "{}"'.format(self._name)


# common grammar elements
_e = pp.CaselessLiteral('e')
_number = (pp.Combine(pp.Word('+-' + pp.nums, pp.nums) +
                      pp.Optional('.' + pp.Optional(pp.Word(pp.nums))) +
                      pp.Optional(_e + pp.Word('+-' + pp.nums, pp.nums)))
           .setResultsName('number'))
_quoted_string = pp.QuotedString('"', '\\').setResultsName('quoted-string')
_identifier = pp.Word(pp.alphas + '_', pp.alphanums + '_').setResultsName('id')
_tph_scope_prefix = (pp.Literal(period.DynScope.TPH.value)
                     .setResultsName('tph-scope-prefix'))
_spc_scope_prefix = (pp.Literal(period.DynScope.SPC.value)
                     .setResultsName('spc-scope-prefix'))
_seh_scope_prefix = (pp.Literal(period.DynScope.SEH.value)
                     .setResultsName('seh-scope-prefix'))
_sec_scope_prefix = (pp.Literal(period.DynScope.SEC.value)
                     .setResultsName('sec-scope-prefix'))
_ec_scope_prefix = (pp.Literal(period.DynScope.EC.value)
                    .setResultsName('ec-scope-prefix'))
_ep_scope_prefix = (pp.Literal(period.DynScope.EP.value)
                    .setResultsName('ep-scope-prefix'))
_dyn_scope_prefix = pp.Group(pp.Group(_tph_scope_prefix |
                                      _spc_scope_prefix |
                                      _seh_scope_prefix |
                                      _sec_scope_prefix |
                                      _ec_scope_prefix |
                                      _ep_scope_prefix) +
                             '.').setResultsName('dyn-scope-prefix')
_parent_scope_prefix = (pp.Group(pp.Literal('$parent') + '.')
                        .setResultsName('parent-scope-prefix'))
_begin_scope_prefix = (pp.Group(pp.Literal('$begin') + '.')
                       .setResultsName('begin-scope-prefix'))
_event_scope_prefix = (pp.Group(pp.Literal('$evt') + '.')
                       .setResultsName('event-scope-prefix'))
_event_field = pp.Group(pp.Optional(_parent_scope_prefix) +
                        pp.Optional(_begin_scope_prefix) +
                        _event_scope_prefix +
                        pp.Optional(_dyn_scope_prefix) +
                        _identifier).setResultsName('event-field')
_event_name = pp.Group(pp.Optional(_parent_scope_prefix) +
                       pp.Optional(_begin_scope_prefix) +
                       _event_scope_prefix +
                       '$name').setResultsName('event-name')
_relop = (pp.Group(pp.Literal('==') | '!=' | '<=' | '>=' | '<' | '>')
          .setResultsName('relop'))
_eqop = pp.Group(pp.Literal('=*') | '==' | '!=').setResultsName('eqop')
_name_comp_expr = pp.Group(_event_name + _eqop +
                           _quoted_string).setResultsName('name-comp-expr')
_number_comp_expr = pp.Group(_event_field + _relop +
                             _number).setResultsName('number-comp-expr')
_string_comp_expr = pp.Group(_event_field + _eqop +
                             _quoted_string).setResultsName('string-comp-expr')
_field_comp_expr = (pp.Group(_event_field.setResultsName('lh') + _relop +
                             _event_field.setResultsName('rh'))
                    .setResultsName('field-comp-expr'))
_comp_expr = (_name_comp_expr |
              _number_comp_expr |
              _string_comp_expr |
              _field_comp_expr)
_not_op = pp.Literal('!').setResultsName('notop')
_and_op = pp.Literal('&&').setResultsName('andop')
_or_op = pp.Literal('||').setResultsName('orop')
_expr = pp.infixNotation(_comp_expr,
                         [
                             (_not_op, 1, pp.opAssoc.RIGHT),
                             (_and_op, 2, pp.opAssoc.LEFT),
                             (_or_op, 2, pp.opAssoc.LEFT)
                         ]).setResultsName('expr')

# period definition grammar elements
_parent_name = pp.Literal('(') + _identifier + ')'
_period_info = (pp.Group(_identifier.setResultsName('name') +
                         (pp.Optional(_parent_name)
                          .setResultsName('parent-name')))
                .setResultsName('period-info'))
_period_def = (pp.Optional(_period_info) + ':' +
               _expr.setResultsName('begin-expr') +
               pp.Optional(pp.Literal(':') +
                           _expr.setResultsName('end-expr')))

# period capture grammar elements
_capture_ref = (pp.Group(pp.Optional(_identifier + '=').setResultsName('var') +
                         (_event_name | _event_field))
                .setResultsName('capture-ref'))
_capture_refs = pp.delimitedList(_capture_ref, ',')
_captures_def = (_identifier.setResultsName('name') + ':' +
                 pp.Optional(_capture_refs.setResultsName('begin-exprs')) +
                 pp.Optional(pp.Literal(':') +
                             _capture_refs.setResultsName('end-exprs')))


# operator string -> function which creates an expression
_OP_TO_EXPR = {
    '=*': lambda lh, rh: period.GlobEq(lh, rh),
    '==': lambda lh, rh: period.Eq(lh, rh),
    '!=': lambda lh, rh: period.LogicalNot(period.Eq(lh, rh)),
    '<': lambda lh, rh: period.Lt(lh, rh),
    '<=': lambda lh, rh: period.LtEq(lh, rh),
    '>': lambda lh, rh: period.Gt(lh, rh),
    '>=': lambda lh, rh: period.GtEq(lh, rh),
}


def _res_to_scope(res):
    if res[-1] == '$name':
        scope = period.EventName()
    elif 'id' in res:
        scope = period.EventFieldName(res['id'])
    else:
        assert(False)

    if 'dyn-scope-prefix' in res:
        dyn_scope = period.DynScope(res['dyn-scope-prefix'][0][0])
        scope = period.DynamicScope(dyn_scope, scope)

    scope = period.EventScope(scope)

    if 'begin-scope-prefix' in res:
        scope = period.BeginScope(scope)

    if 'parent-scope-prefix' in res:
        scope = period.ParentScope(scope)

    return scope


def _res_quoted_string_to_string_expression(res_quoted_string):
    return period.String(str(res_quoted_string))


def _res_number_to_number_expression(res_number):
    return period.Number(float(str(res_number)))


def _create_binary_op(relop, lh, rh):
    return _OP_TO_EXPR[relop[0]](lh, rh)


def _extract_exprs(res):
    exprs = []

    for res_child in res:
        if res_child not in ('&&', '||'):
            expr = _expr_results_to_expression(res_child)
            exprs.append(expr)

    return exprs


def _expr_results_to_expression(res_expr):
    # check for logical op
    if 'notop' in res_expr:
        expr = _expr_results_to_expression(res_expr[1])

        return period.LogicalNot(expr)

    if 'andop' in res_expr:
        exprs = _extract_exprs(res_expr)

        return period.create_conjunction_from_exprs(exprs)

    if 'orop' in res_expr:
        exprs = _extract_exprs(res_expr)

        return period.create_disjunction_from_exprs(exprs)

    res_expr_name = res_expr.getName()

    if res_expr_name == 'name-comp-expr':
        ev_name_expr = _res_to_scope(res_expr['event-name'])
        qstring = res_expr['quoted-string']
        str_expr = _res_quoted_string_to_string_expression(qstring)

        return _create_binary_op(res_expr['eqop'], ev_name_expr, str_expr)

    if res_expr_name == 'number-comp-expr':
        relop = res_expr['relop']
        field_expr = _res_to_scope(res_expr['event-field'])
        number_expr = _res_number_to_number_expression(res_expr['number'])

        return _create_binary_op(relop, field_expr, number_expr)

    if res_expr_name == 'string-comp-expr':
        field_expr = _res_to_scope(res_expr['event-field'])
        qstring = res_expr['quoted-string']
        str_expr = _res_quoted_string_to_string_expression(qstring)

        return _create_binary_op(res_expr['eqop'], field_expr, str_expr)

    if res_expr_name == 'field-comp-expr':
        lh_field_expr = _res_to_scope(res_expr['lh'])
        rh_field_expr = _res_to_scope(res_expr['rh'])

        return _create_binary_op(res_expr['relop'], lh_field_expr,
                                 rh_field_expr)

    assert(False)


def _capture_refs_results_to_captures_exprs(res_capture_refs):
    captures_exprs = {}

    for res_capture_ref in res_capture_refs:
        name = None

        if 'var' in res_capture_ref:
            name = res_capture_ref['var'][0]

        expr = _res_to_scope(res_capture_ref[-1])

        if name is None:
            name = str(expr)

        if name in captures_exprs:
            raise DuplicatePeriodCapture(name)

        captures_exprs[name] = expr

    return captures_exprs


class PeriodDefArgParseResults:
    def __init__(self, parent_name, period_name, begin_expr, end_expr):
        self._parent_name = parent_name
        self._period_name = period_name
        self._begin_expr = begin_expr
        self._end_expr = end_expr

    @property
    def parent_name(self):
        return self._parent_name

    @property
    def period_name(self):
        return self._period_name

    @property
    def begin_expr(self):
        return self._begin_expr

    @property
    def end_expr(self):
        return self._end_expr


class PeriodCapturesDefArgResults:
    def __init__(self, name, begin_captures_exprs, end_captures_exprs):
        self._name = name
        self._begin_captures_exprs = begin_captures_exprs
        self._end_captures_exprs = end_captures_exprs

    @property
    def name(self):
        return self._name

    @property
    def begin_captures_exprs(self):
        return self._begin_captures_exprs

    @property
    def end_captures_exprs(self):
        return self._end_captures_exprs


def parse_period_def_arg(arg):
    try:
        period_def_res = _period_def.parseString(arg.split('/')[-1],
                                                 parseAll=True)
    except Exception:
        raise MalformedExpression(arg)

    period_name = None
    parent_name = None

    if 'period-info' in period_def_res:
        period_info_res = period_def_res['period-info']
        period_name = period_info_res['name']

        if 'parent-name' in period_info_res:
            parent_name = period_info_res['parent-name']['id']

    begin_expr = _expr_results_to_expression(period_def_res['begin-expr'])

    if 'end-expr' in period_def_res:
        end_expr = _expr_results_to_expression(period_def_res['end-expr'])
    else:
        end_expr = begin_expr

    return PeriodDefArgParseResults(parent_name, period_name, begin_expr,
                                    end_expr)


def parse_period_captures_arg(arg):
    try:
        period_captures_res = _captures_def.parseString(arg.split('/')[-1],
                                                        parseAll=True)
    except MalformedExpression:
        raise
    except Exception:
        raise MalformedExpression(arg)

    if 'begin-exprs' in period_captures_res:
        begin_captures_exprs = _capture_refs_results_to_captures_exprs(
            period_captures_res['begin-exprs'])
    else:
        begin_captures_exprs = {}

    if 'end-exprs' in period_captures_res:
        end_captures_exprs = _capture_refs_results_to_captures_exprs(
            period_captures_res['end-exprs'])
    else:
        end_captures_exprs = {}

    return PeriodCapturesDefArgResults(period_captures_res['name'],
                                       begin_captures_exprs,
                                       end_captures_exprs)



# The MIT License (MIT)
#
# Copyright (C) 2016 - Antoine Busque <abusque@efficios.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import unittest
from datetime import date
from lttnganalyses.common import trace_utils
from .utils import TimezoneUtils


# Mock of babeltrace's TraceCollection, used to test date methods
class TraceCollection():
    def __init__(self, begin_ts, end_ts):
        self.begin_ts = begin_ts
        self.end_ts = end_ts

    @property
    def timestamp_begin(self):
        return self.begin_ts

    @property
    def timestamp_end(self):
        return self.end_ts


class TestIsMultiDayTraceCollection(unittest.TestCase):
    def setUp(self):
        self.tz_utils = TimezoneUtils()
        self.tz_utils.set_up_timezone()

    def tearDown(self):
        self.tz_utils.tear_down_timezone()

    def test_same_day(self):
        begin_ts = 683153828123456789
        # 1 hour later
        end_ts = 683157428123456789
        collection = TraceCollection(begin_ts, end_ts)
        result = trace_utils.is_multi_day_trace_collection(collection)

        self.assertFalse(result)

    def test_different_day(self):
        begin_ts = 683153828123456789
        # 24 hours later
        end_ts = 683240228123456789
        collection = TraceCollection(begin_ts, end_ts)
        result = trace_utils.is_multi_day_trace_collection(collection)

        self.assertTrue(result)


class TestGetTraceCollectionDate(unittest.TestCase):
    def setUp(self):
        self.tz_utils = TimezoneUtils()
        self.tz_utils.set_up_timezone()

    def tearDown(self):
        self.tz_utils.tear_down_timezone()

    def test_single_day(self):
        begin_ts = 683153828123456789
        # 1 hour later
        end_ts = 683157428123456789
        collection = TraceCollection(begin_ts, end_ts)
        result = trace_utils.get_trace_collection_date(collection)
        expected = date(1991, 8, 25)

        self.assertEqual(result, expected)

    def test_multi_day(self):
        begin_ts = 683153828123456789
        # 24 hours later
        end_ts = 683240228123456789
        collection = TraceCollection(begin_ts, end_ts)

        self.assertRaises(ValueError, trace_utils.get_trace_collection_date,
                          collection)


class TestGetSyscallName(unittest.TestCase):
    class Event():
        def __init__(self, name):
            self.name = name

    def test_sys(self):
        event = self.Event('sys_open')
        result = trace_utils.get_syscall_name(event)

        self.assertEqual(result, 'open')

    def test_syscall_entry(self):
        event = self.Event('syscall_entry_open')
        result = trace_utils.get_syscall_name(event)

        self.assertEqual(result, 'open')

    def test_not_syscall(self):
        event = self.Event('whatever')

        self.assertRaises(ValueError, trace_utils.get_syscall_name, event)
